#!/usr/bin/env python3

from autovul.base.utils.fim import KFAC, EKFAC
from autovul.base.utils import empty_cache
from autovul.base.utils.logger import MetricLogger, SmoothedValue
from autovul.base.utils.model import accuracy, activate_params
from autovul.base.utils.output import ansi, get_ansi_len, output_iter, prints
from autovul.base.environ import env

import torch
import torch.nn as nn

from tqdm import tqdm

from typing import Union
from collections.abc import Callable
from torch.optim.optimizer import Optimizer
from torch.optim.lr_scheduler import _LRScheduler
import torch.utils.data


def train(module: nn.Module, num_classes: int,
          epoch: int, optimizer: Optimizer, lr_scheduler: _LRScheduler = None,
          grad_clip: float = None, pre_conditioner: Union[KFAC, EKFAC] = None,
          print_prefix: str = 'Epoch', start_epoch: int = 0, resume: int = 0,
          validate_interval: int = 10, save: bool = False, amp: bool = False,
          loader_train: torch.utils.data.DataLoader = None, loader_valid: torch.utils.data.DataLoader = None,
          epoch_fn: Callable[..., None] = None,
          get_data_fn: Callable[..., tuple[torch.Tensor, torch.Tensor]] = None,
          loss_fn: Callable[..., torch.Tensor] = None,
          after_loss_fn: Callable[..., None] = None,
          validate_fn: Callable[..., tuple[float, float]] = None,
          save_fn: Callable[..., None] = None, file_path: str = None, folder_path: str = None, suffix: str = None,
          writer=None, main_tag: str = 'train', tag: str = '',
          accuracy_fn: Callable[..., list[float]] = None,
          verbose: bool = True, indent: int = 0,
          change_train_eval: bool = True, lr_scheduler_freq: str = 'epoch', **kwargs) -> None:
    if epoch <= 0:
        return
    get_data_fn = get_data_fn if get_data_fn is not None else lambda x: x
    loss_fn = loss_fn if loss_fn is not None else nn.CrossEntropyLoss()
    validate_fn = validate_fn if callable(validate_fn) else validate
    accuracy_fn = accuracy_fn if callable(accuracy_fn) else accuracy

    scaler: torch.cuda.amp.GradScaler = None
    if not env['num_gpus']:
        amp = False
    if amp:
        scaler = torch.cuda.amp.GradScaler()
    if validate_interval != 0:
        _, best_acc = validate_fn(loader=loader_valid, get_data_fn=get_data_fn, loss_fn=loss_fn,
                                  writer=None, tag=tag, _epoch=start_epoch,
                                  verbose=verbose, indent=indent, **kwargs)

    params: list[nn.Parameter] = []
    for param_group in optimizer.param_groups:
        params.extend(param_group['params'])
    len_loader_train = len(loader_train)
    total_iter = (epoch - resume) * len_loader_train

    if resume and lr_scheduler:
        for _ in range(resume):
            lr_scheduler.step()
    for _epoch in range(resume, epoch):
        _epoch += 1
        if callable(epoch_fn):
            activate_params(module, [])
            epoch_fn(optimizer=optimizer, lr_scheduler=lr_scheduler,
                     _epoch=_epoch, epoch=epoch, start_epoch=start_epoch)
            activate_params(module, params)
        logger = MetricLogger()
        logger.meters['loss'] = SmoothedValue()
        logger.meters['top1'] = SmoothedValue()
        logger.meters['top5'] = SmoothedValue()
        loader_epoch = loader_train
        if verbose:
            header = '{blue_light}{0}: {1}{reset}'.format(
                print_prefix, output_iter(_epoch, epoch), **ansi)
            header = header.ljust(30 + get_ansi_len(header))
            if env['tqdm']:
                header = '{upline}{clear_line}'.format(**ansi) + header
                loader_epoch = tqdm(loader_epoch)
            loader_epoch = logger.log_every(loader_epoch, header=header, indent=indent)
        if change_train_eval:
            module.train()
        activate_params(module, params)
        optimizer.zero_grad()
        for i, data in enumerate(loader_epoch):
            _iter = _epoch * len_loader_train + i
            # data_time.update(time.perf_counter() - end)
            _input, _label = get_data_fn(data, mode='train')
            if pre_conditioner is not None and not amp:
                pre_conditioner.track.enable()
            _output = module(_input, amp=amp)
            loss = loss_fn(_input, _label, _output=_output, amp=amp)
            if amp:
                scaler.scale(loss).backward()
                if callable(after_loss_fn):
                    after_loss_fn(_input=_input, _label=_label, _output=_output,
                                  loss=loss, optimizer=optimizer, loss_fn=loss_fn,
                                  amp=amp, scaler=scaler,
                                  _iter=_iter, total_iter=total_iter)
                scaler.step(optimizer)
                scaler.update()
            else:
                loss.backward()
                if callable(after_loss_fn):
                    after_loss_fn(_input=_input, _label=_label, _output=_output,
                                  loss=loss, optimizer=optimizer, loss_fn=loss_fn,
                                  amp=amp, scaler=scaler,
                                  _iter=_iter, total_iter=total_iter)
                    # start_epoch=start_epoch, _epoch=_epoch, epoch=epoch)
                if pre_conditioner is not None:
                    pre_conditioner.track.disable()
                    pre_conditioner.step()
                if grad_clip is not None:
                    nn.utils.clip_grad_norm_(params, grad_clip)
                optimizer.step()
            if lr_scheduler and lr_scheduler_freq == 'step':
                lr_scheduler.step()
            optimizer.zero_grad()
            acc1, acc5 = accuracy_fn(_output, _label, num_classes=num_classes, topk=(1, 5))
            batch_size = int(_label.size(0))
            logger.meters['loss'].update(float(loss), batch_size)
            logger.meters['top1'].update(acc1, batch_size)
            logger.meters['top5'].update(acc5, batch_size)
            empty_cache()   # TODO: should it be outside of the dataloader loop?
        if lr_scheduler and lr_scheduler_freq == 'epoch':
            lr_scheduler.step()
        if change_train_eval:
            module.eval()
        activate_params(module, [])
        loss, acc = logger.meters['loss'].global_avg, logger.meters['top1'].global_avg
        if writer is not None:
            from torch.utils.tensorboard import SummaryWriter
            assert isinstance(writer, SummaryWriter)
            writer.add_scalars(main_tag='Loss/' + main_tag, tag_scalar_dict={tag: loss},
                               global_step=_epoch + start_epoch)
            writer.add_scalars(main_tag='Acc/' + main_tag, tag_scalar_dict={tag: acc},
                               global_step=_epoch + start_epoch)
        if validate_interval != 0:
            if _epoch % validate_interval == 0 or _epoch == epoch:
                _, cur_acc = validate_fn(module=module, num_classes=num_classes,
                                         loader=loader_valid, get_data_fn=get_data_fn, loss_fn=loss_fn,
                                         writer=writer, tag=tag, _epoch=_epoch + start_epoch,
                                         verbose=verbose, indent=indent, **kwargs)
                if cur_acc >= best_acc:
                    if verbose:
                        prints('{purple}best result update!{reset}'.format(**ansi), indent=indent)
                        prints(f'Current Acc: {cur_acc:.3f}    Previous Best Acc: {best_acc:.3f}', indent=indent)
                    best_acc = cur_acc
                    if save:
                        save_fn(file_path=file_path, folder_path=folder_path, suffix=suffix, verbose=verbose)
                if verbose:
                    prints('-' * 50, indent=indent)
    module.zero_grad()


def validate(module: nn.Module, num_classes: int, loader: torch.utils.data.DataLoader,
             print_prefix: str = 'Validate', indent: int = 0, verbose: bool = True,
             get_data_fn: Callable[..., tuple[torch.Tensor, torch.Tensor]] = None,
             loss_fn: Callable[..., torch.Tensor] = None,
             writer=None, main_tag: str = 'valid', tag: str = '', _epoch: int = None,
             accuracy_fn: Callable[..., list[float]] = None,
             **kwargs) -> tuple[float, float]:
    module.eval()
    get_data_fn = get_data_fn if get_data_fn is not None else lambda x: x
    loss_fn = loss_fn if loss_fn is not None else nn.CrossEntropyLoss()
    accuracy_fn = accuracy_fn if callable(accuracy_fn) else accuracy
    logger = MetricLogger()
    logger.meters['loss'] = SmoothedValue()
    logger.meters['top1'] = SmoothedValue()
    logger.meters['top5'] = SmoothedValue()
    loader_epoch = loader
    if verbose:
        header = '{yellow}{0}{reset}'.format(print_prefix, **ansi)
        header = header.ljust(max(len(print_prefix), 30) + get_ansi_len(header))
        if env['tqdm']:
            header = '{upline}{clear_line}'.format(**ansi) + header
            loader_epoch = tqdm(loader_epoch)
        loader_epoch = logger.log_every(loader_epoch, header=header, indent=indent)
    for data in loader_epoch:
        _input, _label = get_data_fn(data, mode='valid', **kwargs)
        with torch.no_grad():
            _output = module(_input)
            loss = float(loss_fn(_input, _label, _output=_output, **kwargs))
            acc1, acc5 = accuracy_fn(_output, _label, num_classes=num_classes, topk=(1, 5))
            batch_size = int(_label.size(0))
            logger.meters['loss'].update(loss, batch_size)
            logger.meters['top1'].update(acc1, batch_size)
            logger.meters['top5'].update(acc5, batch_size)
    loss, acc = logger.meters['loss'].global_avg, logger.meters['top1'].global_avg
    if writer is not None and _epoch is not None and main_tag:
        from torch.utils.tensorboard import SummaryWriter
        assert isinstance(writer, SummaryWriter)
        writer.add_scalars(main_tag='Loss/' + main_tag, tag_scalar_dict={tag: loss}, global_step=_epoch)
        writer.add_scalars(main_tag='Acc/' + main_tag, tag_scalar_dict={tag: acc}, global_step=_epoch)
    return loss, acc


def compare(module1: nn.Module, module2: nn.Module, loader: torch.utils.data.DataLoader,
            print_prefix='Validate', indent=0, verbose=True,
            get_data_fn: Callable[..., tuple[torch.Tensor, torch.Tensor]] = None,
            **kwargs) -> float:
    logsoftmax = nn.LogSoftmax(dim=1)
    softmax = nn.Softmax(dim=1)
    module1.eval()
    module2.eval()
    get_data_fn = get_data_fn if get_data_fn is not None else lambda x: x

    def cross_entropy(p: torch.Tensor, q: torch.Tensor) -> torch.Tensor:
        result: torch.Tensor = -softmax(p) * logsoftmax(q)
        return result.sum(1).mean()

    logger = MetricLogger()
    logger.meters['loss'] = SmoothedValue()
    loader_epoch = loader
    if verbose:
        header = '{yellow}{0}{reset}'.format(print_prefix, **ansi)
        header = header.ljust(max(len(print_prefix), 30) + get_ansi_len(header))
        if env['tqdm']:
            header = '{upline}{clear_line}'.format(**ansi) + header
            loader_epoch = tqdm(loader_epoch)
        loader_epoch = logger.log_every(loader_epoch, header=header, indent=indent)
    with torch.no_grad():
        for data in loader_epoch:
            _input, _label = get_data_fn(data, **kwargs)
            _output1, _output2 = module1(_input), module2(_input)
            loss = float(cross_entropy(_output1, _output2))
            batch_size = int(_label.size(0))
            logger.meters['loss'].update(loss, batch_size)
    return logger.meters['loss'].global_avg



#!/usr/bin/env python3
from autovul.vision.models.imagemodel import _ImageModel, ImageModel

import torch.nn as nn
import torchvision.models
from torchvision.models.alexnet import model_urls as urls


class _AlexNet(_ImageModel):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        _model = torchvision.models.alexnet(num_classes=self.num_classes)
        self.features = _model.features
        self.pool = _model.avgpool   # nn.AdaptiveAvgPool2d((6, 6))
        if len(self.classifier) == 1 and isinstance(self.classifier[0], nn.Identity):
            self.classifier = _model.classifier

        # nn.Sequential(
        #     nn.Conv2d(3, 64, kernel_size=11, stride=4, padding=2),
        #     nn.ReLU(inplace=True),
        #     nn.MaxPool2d(kernel_size=3, stride=2),
        #     nn.Conv2d(64, 192, kernel_size=5, padding=2),
        #     nn.ReLU(inplace=True),
        #     nn.MaxPool2d(kernel_size=3, stride=2),
        #     nn.Conv2d(192, 384, kernel_size=3, padding=1),
        #     nn.ReLU(inplace=True),
        #     nn.Conv2d(384, 256, kernel_size=3, padding=1),
        #     nn.ReLU(inplace=True),
        #     nn.Conv2d(256, 256, kernel_size=3, padding=1),
        #     nn.ReLU(inplace=True),
        #     nn.MaxPool2d(kernel_size=3, stride=2),
        # )

        # nn.Sequential(
        #     nn.Dropout(),
        #     nn.Linear(256 * 6 * 6, 4096),
        #     nn.ReLU(inplace=True),
        #     nn.Dropout(),
        #     nn.Linear(4096, 4096),
        #     nn.ReLU(inplace=True),
        #     nn.Linear(4096, num_classes),
        # )


class AlexNet(ImageModel):
    available_models = ['alexnet']
    model_urls = urls

    def __init__(self, name: str = 'alexnet', model: type[_AlexNet] = _AlexNet, **kwargs):
        super().__init__(name=name, model=model, **kwargs)



#!/usr/bin/env python3

from autovul.vision.models.imagemodel import _ImageModel, ImageModel
from autovul.vision.utils.model_archs.pnasnet import PNASNetA, PNASNetB

import torch.nn as nn
from collections import OrderedDict


class _PNASNet(_ImageModel):

    def __init__(self, cell_type: str = 'b', **kwargs):
        super().__init__(**kwargs)
        assert cell_type in ['a', 'b'], cell_type
        ModelClass = PNASNetA if cell_type == 'a' else PNASNetB
        _model = ModelClass(num_classes=self.num_classes)
        self.features = nn.Sequential(OrderedDict([
            ('conv1', _model.conv1),
            ('bn1', _model.bn1),
            ('relu', nn.ReLU(inplace=True)),
            ('layer1', _model.layer1),
            ('layer2', _model.layer2),
            ('layer3', _model.layer3),
            ('layer4', _model.layer4),
            ('layer5', _model.layer5)
        ]))
        # self.pool = nn.AvgPool2d(8)
        self.classifier = nn.Sequential(OrderedDict([
            ('fc', _model.linear)
        ]))


class PNASNet(ImageModel):
    available_models = ['pnasnet', 'pnasnet_a', 'pnasnet_b']

    def __init__(self, name: str = 'pnasnet', cell_type: str = 'b',
                 model: type[_PNASNet] = _PNASNet, **kwargs):
        if name == 'pnasnet':
            name += f'_{cell_type}'
        else:
            assert name in ['pnasnet_a', 'pnasnet_b'], name
            cell_type = 'a' if name == 'pnasnet_a' else 'b'
        super().__init__(name=name, cell_type=cell_type, model=model, **kwargs)



#!/usr/bin/env python3

from .genotypes import PRIMITIVES
from .operations import get_op

import torch
import torch.nn as nn
import torch.nn.functional as F


class MixedOp(nn.Module):
    """ Mixed operation """

    def __init__(self, C: int, stride: int):
        super().__init__()
        self._ops = nn.ModuleList()
        for primitive in PRIMITIVES:
            op = get_op(primitive, C, stride, affine=False)
            self._ops.append(op)

    def forward(self, x: torch.Tensor, weights: torch.Tensor) -> torch.Tensor:
        """
        Args:
            x: input
            weights: weight for each operation
        """
        return sum(w * op(x) for w, op in zip(weights, self._ops))


class Cell(nn.Module):
    def __init__(self, steps: int, multiplier: int,
                 C_prev_prev: int, C_prev: int, C: int,
                 reduction: bool, reduction_prev: bool):
        super().__init__()
        self.reduction = reduction

        op_name = 'factorized_reduce' if reduction_prev else 'conv'
        self.preprocess0 = get_op(op_name, C_prev_prev, affine=False, C_out=C)
        self.preprocess1 = get_op('conv', C_prev, affine=False, C_out=C)

        self._steps = steps
        self._multiplier = multiplier

        self._ops = nn.ModuleList()
        for i in range(self._steps):
            for j in range(2 + i):
                stride = 2 if reduction and j < 2 else 1
                op = MixedOp(C, stride)
                self._ops.append(op)

    def forward(self, s0: torch.Tensor, s1: torch.Tensor, weights: torch.Tensor) -> torch.Tensor:
        s0 = self.preprocess0(s0)
        s1 = self.preprocess1(s1)
        states = [s0, s1]
        offset = 0
        for i in range(self._steps):
            s = sum(self._ops[offset + j](h, weights[offset + j]) for j, h in enumerate(states))
            offset += len(states)
            states.append(s)
        return torch.cat(states[-self._multiplier:], dim=1)


class Feature_Extractor(nn.Module):
    def __init__(self, C: int, layers: int,
                 steps: int = 4, multiplier: int = 4, stem_multiplier: int = 3):
        super().__init__()
        self._C = C
        self._layers = layers
        self._steps = steps
        self._multiplier = multiplier

        C_curr = stem_multiplier * C
        self.stem = nn.Sequential(
            nn.Conv2d(3, C_curr, 3, padding=1, bias=False),
            nn.BatchNorm2d(C_curr)
        )

        self.cells = nn.ModuleList()
        C_prev_prev, C_prev, C_curr = C_curr, C_curr, C
        reduction_prev = False
        for i in range(layers):
            if i in [layers // 3, 2 * layers // 3]:
                C_curr *= 2
                reduction = True
            else:
                reduction = False
            cell = Cell(steps, multiplier, C_prev_prev, C_prev, C_curr, reduction, reduction_prev)
            reduction_prev = reduction
            self.cells.append(cell)
            C_prev_prev, C_prev = C_prev, multiplier * C_curr

    def forward(self, input):
        s0 = s1 = self.stem(input)
        for i, cell in enumerate(self.cells):
            cell: Cell
            if cell.reduction:
                weights = F.softmax(self.alphas_reduce, dim=-1)
            else:
                weights = F.softmax(self.alphas_normal, dim=-1)
            s0, s1 = s1, cell(s0, s1, weights)
        return s1



#!/usr/bin/env python3

import autovul.vision
from autovul.vision.attacks import PoisonRandom

from autovul.vision.utils import summary
import argparse

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    autovul.vision.environ.add_argument(parser)
    autovul.vision.datasets.add_argument(parser)
    autovul.vision.models.add_argument(parser)
    autovul.vision.trainer.add_argument(parser)
    autovul.vision.attacks.add_argument(parser)
    args = parser.parse_args()

    env = autovul.vision.environ.create(**args.__dict__)
    dataset = autovul.vision.datasets.create(**args.__dict__)
    model = autovul.vision.models.create(dataset=dataset, **args.__dict__)
    trainer = autovul.vision.trainer.create(dataset=dataset, model=model, **args.__dict__)
    attack: PoisonRandom = autovul.vision.attacks.create(dataset=dataset, model=model, **args.__dict__)

    if env['verbose']:
        summary(env=env, dataset=dataset, model=model, train=trainer, attack=attack)
    attack.load()
    attack.validate_fn()



#!/usr/bin/env python3

# https://github.com/facebookresearch/GradientEpisodicMemory/blob/master/model/common.py

# Copyright 2017-present, Facebook, Inc.
# All rights reserved.
#
# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

import torch
import torch.nn as nn
from torchvision.models.resnet import conv1x1, conv3x3, ResNet

from typing import Any, Callable, Optional, Union


class BasicBlock_AP(nn.Module):
    expansion: int = 1

    def __init__(
        self,
        inplanes: int,
        planes: int,
        stride: int = 1,
        downsample: Optional[nn.Module] = None,
        groups: int = 1,
        base_width: int = 64,
        dilation: int = 1,
        norm_layer: Optional[Callable[..., nn.Module]] = None
    ) -> None:
        super().__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        if groups != 1 or base_width != 64:
            raise ValueError('BasicBlock only supports groups=1 and base_width=64')
        if dilation > 1:
            raise NotImplementedError("Dilation > 1 not supported in BasicBlock")
        # Both self.conv1 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv3x3(inplanes, planes, stride=1)    # modification
        self.bn1 = norm_layer(planes)
        self.relu = nn.ReLU(inplace=True)

        self.pool = nn.Identity() if stride == 1 else nn.AvgPool2d(kernel_size=stride, stride=stride)   # modification

        self.conv2 = conv3x3(planes, planes)
        self.bn2 = norm_layer(planes)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.pool(out)    # modification

        out = self.conv2(out)
        out = self.bn2(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class Bottleneck_AP(nn.Module):
    # Bottleneck in torchvision places the stride for downsampling at 3x3 convolution(self.conv2)
    # while original implementation places the stride at the first 1x1 convolution(self.conv1)
    # according to "Deep residual learning for image recognition"https://arxiv.org/abs/1512.03385.
    # This variant is also known as ResNet V1.5 and improves accuracy according to
    # https://ngc.nvidia.com/catalog/model-scripts/nvidia:resnet_50_v1_5_for_pytorch.

    expansion: int = 4

    def __init__(
        self,
        inplanes: int,
        planes: int,
        stride: int = 1,
        downsample: Optional[nn.Module] = None,
        groups: int = 1,
        base_width: int = 64,
        dilation: int = 1,
        norm_layer: Optional[Callable[..., nn.Module]] = None
    ) -> None:
        super().__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        width = int(planes * (base_width / 64.)) * groups
        # Both self.conv2 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv1x1(inplanes, width)
        self.bn1 = norm_layer(width)
        self.conv2 = conv3x3(width, width, stride=1, groups=groups, dilation=dilation)  # modification
        self.bn2 = norm_layer(width)
        self.pool = nn.Identity() if stride == 1 else nn.AvgPool2d(kernel_size=stride, stride=stride)   # modification
        self.conv3 = conv1x1(width, planes * self.expansion)
        self.bn3 = norm_layer(planes * self.expansion)
        self.relu = nn.ReLU(inplace=True)
        self.downsample = downsample
        self.stride = stride

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.pool(out)    # modification

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        out = self.relu(out)

        return out


class ResNet_AP(ResNet):
    def __init__(self, *args, pool_size: int = 4, **kwargs):
        super().__init__(*args, **kwargs)
        self.avgpool = nn.Identity()    # modification
        self.fc = nn.Linear(self.fc.in_features * (pool_size * pool_size), self.fc.out_features)    # modification

    def _make_layer(self, block: type[Union[BasicBlock_AP, Bottleneck_AP]], planes: int, blocks: int,
                    stride: int = 1, dilate: bool = False) -> nn.Sequential:
        norm_layer = self._norm_layer
        downsample = None
        previous_dilation = self.dilation
        if dilate:
            self.dilation *= stride
            stride = 1
        if stride != 1 or self.inplanes != planes * block.expansion:
            downsample = nn.Sequential(
                conv1x1(self.inplanes, planes * block.expansion, stride=1),  # modification
                nn.AvgPool2d(kernel_size=stride, stride=stride),    # modification
                norm_layer(planes * block.expansion),
            )

        layers = []
        layers.append(block(self.inplanes, planes, stride, downsample, self.groups,
                            self.base_width, previous_dilation, norm_layer))
        self.inplanes = planes * block.expansion
        for _ in range(1, blocks):
            layers.append(block(self.inplanes, planes, groups=self.groups,
                                base_width=self.base_width, dilation=self.dilation,
                                norm_layer=norm_layer))

        return nn.Sequential(*layers)


def _resnet(
    block: type[Union[BasicBlock_AP, Bottleneck_AP]],
    layers: list[int],
    **kwargs: Any
) -> ResNet_AP:
    return ResNet_AP(block, layers, **kwargs)


def resnet18(**kwargs: Any) -> ResNet_AP:
    r"""ResNet-18 model from
    `"Deep Residual Learning for Image Recognition" <https://arxiv.org/pdf/1512.03385.pdf>`_.

    """
    return _resnet(BasicBlock_AP, [2, 2, 2, 2], **kwargs)



#!/usr/bin/env python3

from .adv import *
from .poison import *
from .backdoor import *

from autovul.vision.configs import Config, config
from autovul.vision.datasets import ImageSet
import autovul.base.attacks
from autovul.base.attacks import Attack
import argparse
from typing import Union

class_dict = {
    # adversarial attack
    'pgd': PGD,

    # poisoning attack
    'poison_basic': PoisonBasic,
    'poison_random': PoisonRandom,

    # backdoor attack
    'badnet': BadNet,
    'trojannn': TrojanNN,
    'latent_backdoor': LatentBackdoor,
    'imc': IMC,
}


def add_argument(parser: argparse.ArgumentParser, attack_name: str = None, attack: Union[str, Attack] = None,
                 class_dict: dict[str, type[Attack]] = class_dict):
    return autovul.base.attacks.add_argument(parser=parser, attack_name=attack_name, attack=attack,
                                             class_dict=class_dict)


def create(attack_name: str = None, attack: Union[str, Attack] = None,
           dataset_name: str = None, dataset: Union[str, ImageSet] = None,
           config: Config = config, class_dict: dict[str, type[Attack]] = class_dict, **kwargs):
    return autovul.base.attacks.create(attack_name=attack_name, attack=attack,
                                       dataset_name=dataset_name, dataset=dataset,
                                       config=config, class_dict=class_dict, **kwargs)



#!/usr/bin/env python3

from .output import ansi
from .param import Param
from .others import get_name
from autovul.base.configs import config

import torch
# import torch.distributed as dist
import numpy as np
import random

from typing import TYPE_CHECKING
from typing import Union    # TODO: python 3.10
from autovul.base.configs import Config
import argparse
if TYPE_CHECKING:
    import torch.backends.cudnn


class Env(Param):

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        group.add_argument('--config', dest='config_path',
                           help='cmd config file path. (``package < project < cmd_config < cmd_param``)')

        group.add_argument('--seed', type=int,
                           help='the random seed for numpy, torch and cuda, defaults to config[env][seed]=1228')
        group.add_argument('--cache_threshold', type=float,
                           help='the threshold (MB) to call torch.cuda.empty_cache(), defaults to config[env][cache_threshold]=None (never).')

        group.add_argument('--device', help='set to "cpu" to force cpu-only '
                           'and "gpu", "cuda" for gpu-only, defaults to None.')
        group.add_argument('--benchmark', action='store_true',
                           help='use torch.backends.cudnn.benchmark to accelerate without deterministic, defaults to False.')
        group.add_argument('--verbose', type=int, default=0,
                           help='show arguments and module information, defaults to 0.')
        group.add_argument('--color', action='store_true', help='colorful Output.')
        group.add_argument('--tqdm', action='store_true', help='show tqdm progress bar.')

        # group.add_argument('--world_size', type=int, default=1)
        return group


env = Env(default=None)


def add_argument(parser: argparse.ArgumentParser):
    group = parser.add_argument_group('{yellow}env{reset}'.format(**ansi))
    env.add_argument(group)
    return group


def create(config_path: str = None, dataset_name: str = None, dataset: str = None,
           seed: int = None, benchmark: bool = None,
           config: Config = config,
           cache_threshold: float = None, verbose: int = None,
           color: bool = None, tqdm: bool = None, **kwargs) -> Env:
    other_kwargs = {'cache_threshold': cache_threshold, 'verbose': verbose, 'color': color, 'tqdm': tqdm}
    config.update_cmd(config_path)
    dataset_name = get_name(name=dataset_name, module=dataset, arg_list=['-d', '--dataset'])
    dataset_name = dataset_name if dataset_name is not None else config.get_full_config()['dataset']['default_dataset']
    result = config.get_config(dataset_name=dataset_name)['env'].update(other_kwargs)
    env.update(config_path=config_path, **result)
    ansi.switch(env['color'])
    if seed is None and 'seed' in env.keys():
        seed = env['seed']
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)

    num_gpus: int = torch.cuda.device_count()
    device: Union[str, int] = result['device']
    if device == 'none':
        device = None
    else:
        if device is None or device == 'auto':
            device = 'cuda' if num_gpus else 'cpu'
        if isinstance(device, (str, int)):
            device = torch.device(device)
        if device.type == 'cpu':
            num_gpus = 0
        if device.index is not None and torch.cuda.is_available():
            num_gpus = 1
    if num_gpus == 0:
        device = torch.device('cpu')
    if benchmark is None and 'benchmark' in env.keys():
        benchmark = env['benchmark']
    if benchmark:
        torch.backends.cudnn.benchmark = benchmark

    # dist_backend = 'nccl' if num_gpus and dist.is_nccl_available() else 'gloo'
    # dist.init_process_group(backend=dist_backend)
    env.update(seed=seed, device=device, benchmark=benchmark, num_gpus=num_gpus)
    return env


create()


# def init_distributed_mode():



#!/usr/bin/env python3

from .tensor import to_list

import torch
import numpy as np

from typing import TYPE_CHECKING
from typing import Union    # TODO: python 3.10
if TYPE_CHECKING:
    import torch.utils.data


class TensorListDataset(torch.utils.data.Dataset):
    def __init__(self, data: torch.Tensor = None, targets: list[int] = None, **kwargs):
        super().__init__(**kwargs)
        self.data = data
        self.targets = to_list(targets)
        assert len(self.data) == len(self.targets)

    def __getitem__(self, index: Union[int, slice]) -> tuple[torch.Tensor, int]:
        return self.data[index], self.targets[index]

    def __len__(self):
        return len(self.targets)


class IndexDataset(torch.utils.data.Dataset):
    def __init__(self, dataset: torch.utils.data.Dataset, indices: list[int] = None, **kwargs):
        super().__init__(**kwargs)
        self.dataset = dataset
        if indices is None:
            indices = list(range(len(self.dataset)))
        self.indices = indices

    def __getitem__(self, index: Union[int, slice]) -> tuple:
        return (*self.dataset[index], self.indices[index])

    def __len__(self):
        return len(self.dataset)


def dataset_to_list(dataset: torch.utils.data.Dataset, label_only: bool = False,
                    force: bool = True, shuffle: bool = False) -> tuple[list, list[int]]:
    if not force:
        if label_only and 'targets' in dataset.__dict__.keys():
            return None, list(dataset.targets)
        if 'data' in dataset.__dict__.keys() and 'targets' in dataset.__dict__.keys():
            data = dataset.data
            if isinstance(data, np.ndarray):
                data = torch.as_tensor(data)
            if isinstance(data, torch.Tensor):
                if data.max() > 2:
                    data = data.to(dtype=torch.float) / 255
                data = [img for img in data]
            return data, list(dataset.targets)
    data, targets = list(zip(*dataset))[:2]
    if label_only:
        data = None
    else:
        data = list(data)
    targets = list(targets)
    return data, targets


def shuffle_idx(len: int, seed: int = None) -> np.ndarray:
    idx_arr: np.ndarray = np.arange(len)
    if seed is not None:
        np.random.seed(seed)
    np.random.shuffle(idx_arr)


def sample_batch(dataset: torch.utils.data.Dataset, batch_size: int = None,
                 idx: list[int] = None) -> tuple[list, list[int]]:
    if idx is None:
        assert len(dataset) >= batch_size
        idx = torch.randperm(len(dataset))[:batch_size]
    else:
        assert len(dataset) > max(idx)
    subset = torch.utils.data.Subset(dataset, idx)
    return dataset_to_list(subset)


def split_dataset(dataset: Union[torch.utils.data.Dataset, torch.utils.data.Subset],
                  length: int = None, percent=None, shuffle: bool = True, seed: int = None
                  ) -> tuple[torch.utils.data.Subset, torch.utils.data.Subset]:
    assert (length is None) != (percent is None)  # XOR check
    length = length if length is not None else int(len(dataset) * percent)
    indices = np.arange(len(dataset))
    if shuffle:
        if seed is not None:
            np.random.seed(seed)
        np.random.shuffle(indices)
    if isinstance(dataset, torch.utils.data.Subset):
        idx = np.array(dataset.indices)
        indices = idx[indices]
        dataset = dataset.dataset
    subset1 = torch.utils.data.Subset(dataset, indices[:length])
    subset2 = torch.utils.data.Subset(dataset, indices[length:])
    return subset1, subset2


def get_class_subset(dataset: torch.utils.data.Dataset, class_list: Union[int, list[int]]) -> torch.utils.data.Subset:
    class_list = [class_list] if isinstance(class_list, int) else class_list
    indices = np.arange(len(dataset))
    if isinstance(dataset, torch.utils.data.Subset):
        idx = np.array(dataset.indices)
        indices = idx[indices]
        dataset = dataset.dataset
    _, targets = dataset_to_list(dataset=dataset, label_only=True)
    idx_bool = np.isin(targets, class_list)
    idx = np.arange(len(dataset))[idx_bool]
    idx = np.intersect1d(idx, indices)
    return torch.utils.data.Subset(dataset, idx)



#!/usr/bin/env python3
from autovul.vision.models.imagemodel import _ImageModel, ImageModel

import torch
import torch.nn as nn
import torchvision.models
from torchvision.models.shufflenetv2 import model_urls as urls
from collections import OrderedDict
from collections.abc import Callable


class _ShuffleNetV2(_ImageModel):
    def __init__(self, name: str = 'shufflenetv2_x1.0', **kwargs):
        try:
            sub_type: str = name[13:]
            assert sub_type in ['x0.5', 'x1.0', 'x1.5', 'x2.0',
                                'x0.5_comp', 'x1.0_comp', 'x1.5_comp', 'x2.0_comp'], f'{name=}'
        except Exception:
            raise AssertionError("model name should be in ["
                                 "'shufflenetv2_x0.5', 'shufflenetv2_x1.0', "
                                 "'shufflenetv2_x1.5', 'shufflenetv2_x2.0', "
                                 "'shufflenetv2_x0.5_comp', 'shufflenetv2_x1.0_comp', "
                                 "'shufflenetv2_x1.5_comp', 'shufflenetv2_x2.0_comp']")
        super().__init__(**kwargs)
        arch = sub_type[1:4].replace('.', '_')
        ModelClass: Callable[..., torchvision.models.ShuffleNetV2] = getattr(
            torchvision.models, f'shufflenet_v2_x{arch}')
        _model = ModelClass(num_classes=self.num_classes)
        module_list: list[nn.Module] = []
        if 'comp' in sub_type:
            conv1: nn.Conv2d = _model.conv1[0]
            _model.conv1[0] = nn.Conv2d(conv1.in_channels, conv1.out_channels, kernel_size=conv1.kernel_size,
                                        stride=1, padding=conv1.padding, dilation=conv1.dilation,
                                        groups=conv1.groups, bias=False)
            module_list.append(('conv1', _model.conv1))
        else:
            module_list.append(('conv1', _model.conv1))
            module_list.append(('maxpool', _model.maxpool))
        module_list.extend([
            ('stage2', _model.stage2),
            ('stage3', _model.stage3),
            ('stage4', _model.stage4),
            ('conv5', _model.conv5),
        ])
        self.features = nn.Sequential(OrderedDict(module_list))
        self.classifier = nn.Sequential(OrderedDict([
            ('fc', _model.fc)
        ]))


class ShuffleNetV2(ImageModel):
    available_models = ['shufflenetv2',
                        'shufflenetv2_x0.5', 'shufflenetv2_x1.0', 'shufflenetv2_x1.5', 'shufflenetv2_x2.0',
                        'shufflenetv2_x0.5_comp', 'shufflenetv2_x1.0_comp', 'shufflenetv2_x1.5_comp', 'shufflenetv2_x2.0_comp', ]

    model_urls = urls

    def __init__(self, name: str = 'shufflenetv2', model: type[_ShuffleNetV2] = _ShuffleNetV2, **kwargs):
        super().__init__(name=name, model=model, **kwargs)

    def get_official_weights(self, **kwargs) -> OrderedDict[str, torch.Tensor]:
        _dict = super().get_official_weights(**kwargs)
        new_dict = OrderedDict()
        for i, (key, value) in enumerate(_dict.items()):
            prefix = 'features.' if i < len(_dict) - 2 else 'classifier.'
            new_dict[prefix + key] = value
        return new_dict



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import logging
import warnings
from collections import OrderedDict

import torch.nn as nn


logger = logging.getLogger(__name__)


_counter = 0


def global_mutable_counting():
    """
    A program level counter starting from 1.
    """
    global _counter
    _counter += 1
    return _counter


class Mutable(nn.Module):
    """
    Mutable is designed to function as a normal layer, with all necessary operators' weights.
    States and weights of architectures should be included in mutator, instead of the layer itself.

    Mutable has a key, which marks the identity of the mutable. This key can be used by users to share
    decisions among different mutables. In mutator's implementation, mutators should use the key to
    distinguish different mutables. Mutables that share the same key should be "similar" to each other.

    Currently the default scope for keys is global. By default, the keys uses a global counter from 1 to
    produce unique ids.

    Parameters
    ----------
    key : str
        The key of mutable.

    Notes
    -----
    The counter is program level, but mutables are model level. In case multiple models are defined, and
    you want to have `counter` starting from 1 in the second model, it's recommended to assign keys manually
    instead of using automatic keys.
    """

    def __init__(self, key=None):
        super().__init__()
        if key is not None:
            if not isinstance(key, str):
                key = str(key)
                logger.warning("Warning: key \"%s\" is not string, converted to string.", key)
            self._key = key
        else:
            self._key = self.__class__.__name__ + str(global_mutable_counting())
        self.init_hook = self.forward_hook = None

    def __deepcopy__(self, memodict=None):
        raise NotImplementedError("Deep copy doesn't work for mutables.")

    def __call__(self, *args, **kwargs):
        self._check_built()
        return super().__call__(*args, **kwargs)

    def set_mutator(self, mutator):
        if "mutator" in self.__dict__:
            raise RuntimeError("`set_mutator` is called more than once. Did you parse the search space multiple times? "
                               "Or did you apply multiple fixed architectures?")
        self.__dict__["mutator"] = mutator

    @property
    def key(self):
        """
        Read-only property of key.
        """
        return self._key

    @property
    def name(self):
        """
        After the search space is parsed, it will be the module name of the mutable.
        """
        return self._name if hasattr(self, "_name") else "_key"

    @name.setter
    def name(self, name):
        self._name = name

    def _check_built(self):
        if not hasattr(self, "mutator"):
            raise ValueError(
                "Mutator not set for {}. You might have forgotten to initialize and apply your mutator. "
                "Or did you initialize a mutable on the fly in forward pass? Move to `__init__` "
                "so that trainer can locate all your mutables. See NNI docs for more details.".format(self))


class MutableScope(Mutable):
    """
    Mutable scope marks a subgraph/submodule to help mutators make better decisions.

    If not annotated with mutable scope, search space will be flattened as a list. However, some mutators might
    need to leverage the concept of a "cell". So if a module is defined as a mutable scope, everything in it will
    look like "sub-search-space" in the scope. Scopes can be nested.

    There are two ways mutators can use mutable scope. One is to traverse the search space as a tree during initialization
    and reset. The other is to implement `enter_mutable_scope` and `exit_mutable_scope`. They are called before and after
    the forward method of the class inheriting mutable scope.

    Mutable scopes are also mutables that are listed in the mutator.mutables (search space), but they are not supposed
    to appear in the dict of choices.

    Parameters
    ----------
    key : str
        Key of mutable scope.
    """

    def __init__(self, key):
        super().__init__(key=key)

    def _check_built(self):
        return True  # bypass the test because it's deprecated

    def __call__(self, *args, **kwargs):
        if not hasattr(self, 'mutator'):
            return super().__call__(*args, **kwargs)
        warnings.warn("`MutableScope` is deprecated in Retiarii.", DeprecationWarning)
        try:
            self._check_built()
            self.mutator.enter_mutable_scope(self)
            return super().__call__(*args, **kwargs)
        finally:
            self.mutator.exit_mutable_scope(self)


class LayerChoice(Mutable):
    """
    Layer choice selects one of the ``op_candidates``, then apply it on inputs and return results.
    In rare cases, it can also select zero or many.

    Layer choice does not allow itself to be nested.

    Parameters
    ----------
    op_candidates : list of nn.Module or OrderedDict
        A module list to be selected from.
    reduction : str
        ``mean``, ``concat``, ``sum`` or ``none``. Policy if multiples are selected.
        If ``none``, a list is returned. ``mean`` returns the average. ``sum`` returns the sum.
        ``concat`` concatenate the list at dimension 1.
    return_mask : bool
        If ``return_mask``, return output tensor and a mask. Otherwise return tensor only.
    key : str
        Key of the input choice.

    Attributes
    ----------
    length : int
        Deprecated. Number of ops to choose from. ``len(layer_choice)`` is recommended.
    names : list of str
        Names of candidates.
    choices : list of Module
        Deprecated. A list of all candidate modules in the layer choice module.
        ``list(layer_choice)`` is recommended, which will serve the same purpose.

    Notes
    -----
    ``op_candidates`` can be a list of modules or a ordered dict of named modules, for example,

    .. code-block:: python

        self.op_choice = LayerChoice(OrderedDict([
            ("conv3x3", nn.Conv2d(3, 16, 128)),
            ("conv5x5", nn.Conv2d(5, 16, 128)),
            ("conv7x7", nn.Conv2d(7, 16, 128))
        ]))

    Elements in layer choice can be modified or deleted. Use ``del self.op_choice["conv5x5"]`` or
    ``self.op_choice[1] = nn.Conv3d(...)``. Adding more choices is not supported yet.
    """

    def __init__(self, op_candidates, reduction="sum", return_mask=False, key=None):
        super().__init__(key=key)
        self.names = []
        if isinstance(op_candidates, OrderedDict):
            for name, module in op_candidates.items():
                assert name not in ["length", "reduction", "return_mask", "_key", "key", "names"], \
                    "Please don't use a reserved name '{}' for your module.".format(name)
                self.add_module(name, module)
                self.names.append(name)
        elif isinstance(op_candidates, list):
            for i, module in enumerate(op_candidates):
                self.add_module(str(i), module)
                self.names.append(str(i))
        else:
            raise TypeError("Unsupported op_candidates type: {}".format(type(op_candidates)))
        self.reduction = reduction
        self.return_mask = return_mask

    def __getitem__(self, idx):
        if isinstance(idx, str):
            return self._modules[idx]
        return list(self)[idx]

    def __setitem__(self, idx, module):
        key = idx if isinstance(idx, str) else self.names[idx]
        return setattr(self, key, module)

    def __delitem__(self, idx):
        if isinstance(idx, slice):
            for key in self.names[idx]:
                delattr(self, key)
        else:
            if isinstance(idx, str):
                key, idx = idx, self.names.index(idx)
            else:
                key = self.names[idx]
            delattr(self, key)
        del self.names[idx]

    @property
    def length(self):
        warnings.warn("layer_choice.length is deprecated. Use `len(layer_choice)` instead.", DeprecationWarning)
        return len(self)

    def __len__(self):
        return len(self.names)

    def __iter__(self):
        return map(lambda name: self._modules[name], self.names)

    @property
    def choices(self):
        warnings.warn("layer_choice.choices is deprecated. Use `list(layer_choice)` instead.", DeprecationWarning)
        return list(self)

    def forward(self, *args, **kwargs):
        """
        Returns
        -------
        tuple of tensors
            Output and selection mask. If ``return_mask`` is ``False``, only output is returned.
        """
        out, mask = self.mutator.on_forward_layer_choice(self, *args, **kwargs)
        if self.return_mask:
            return out, mask
        return out


class InputChoice(Mutable):
    """
    Input choice selects ``n_chosen`` inputs from ``choose_from`` (contains ``n_candidates`` keys). For beginners,
    use ``n_candidates`` instead of ``choose_from`` is a safe option. To get the most power out of it, you might want to
    know about ``choose_from``.

    The keys in ``choose_from`` can be keys that appear in past mutables, or ``NO_KEY`` if there are no suitable ones.
    The keys are designed to be the keys of the sources. To help mutators make better decisions,
    mutators might be interested in how the tensors to choose from come into place. For example, the tensor is the
    output of some operator, some node, some cell, or some module. If this operator happens to be a mutable (e.g.,
    ``LayerChoice`` or ``InputChoice``), it has a key naturally that can be used as a source key. If it's a
    module/submodule, it needs to be annotated with a key: that's where a :class:`MutableScope` is needed.

    In the example below, ``input_choice`` is a 4-choose-any. The first 3 is semantically output of cell1, output of cell2,
    output of cell3 with respectively. Notice that an extra max pooling is followed by cell1, indicating x1 is not
    "actually" the direct output of cell1.

    .. code-block:: python

        class Cell(MutableScope):
            pass

        class Net(nn.Module):
            def __init__(self):
                self.cell1 = Cell("cell1")
                self.cell2 = Cell("cell2")
                self.op = LayerChoice([conv3x3(), conv5x5()], key="op")
                self.input_choice = InputChoice(choose_from=["cell1", "cell2", "op", InputChoice.NO_KEY])

            def forward(self, x):
                x1 = max_pooling(self.cell1(x))
                x2 = self.cell2(x)
                x3 = self.op(x)
                x4 = torch.zeros_like(x)
                return self.input_choice([x1, x2, x3, x4])

    Parameters
    ----------
    n_candidates : int
        Number of inputs to choose from.
    choose_from : list of str
        List of source keys to choose from. At least of one of ``choose_from`` and ``n_candidates`` must be fulfilled.
        If ``n_candidates`` has a value but ``choose_from`` is None, it will be automatically treated as ``n_candidates``
        number of empty string.
    n_chosen : int
        Recommended inputs to choose. If None, mutator is instructed to select any.
    reduction : str
        ``mean``, ``concat``, ``sum`` or ``none``. See :class:`LayerChoice`.
    return_mask : bool
        If ``return_mask``, return output tensor and a mask. Otherwise return tensor only.
    key : str
        Key of the input choice.
    """

    NO_KEY = ""

    def __init__(self, n_candidates=None, choose_from=None, n_chosen=None,
                 reduction="sum", return_mask=False, key=None):
        super().__init__(key=key)
        # precondition check
        assert n_candidates is not None or choose_from is not None, "At least one of `n_candidates` and `choose_from`" \
                                                                    "must be not None."
        if choose_from is not None and n_candidates is None:
            n_candidates = len(choose_from)
        elif choose_from is None and n_candidates is not None:
            choose_from = [self.NO_KEY] * n_candidates
        assert n_candidates == len(choose_from), "Number of candidates must be equal to the length of `choose_from`."
        assert n_candidates > 0, "Number of candidates must be greater than 0."
        assert n_chosen is None or 0 <= n_chosen <= n_candidates, "Expected selected number must be None or no more " \
                                                                  "than number of candidates."

        self.n_candidates = n_candidates
        self.choose_from = choose_from.copy()
        self.n_chosen = n_chosen
        self.reduction = reduction
        self.return_mask = return_mask

    def forward(self, optional_inputs):
        """
        Forward method of LayerChoice.

        Parameters
        ----------
        optional_inputs : list or dict
            Recommended to be a dict. As a dict, inputs will be converted to a list that follows the order of
            ``choose_from`` in initialization. As a list, inputs must follow the semantic order that is the same as
            ``choose_from``.

        Returns
        -------
        tuple of tensors
            Output and selection mask. If ``return_mask`` is ``False``, only output is returned.
        """
        optional_input_list = optional_inputs
        if isinstance(optional_inputs, dict):
            optional_input_list = [optional_inputs[tag] for tag in self.choose_from]
        assert isinstance(optional_input_list, list), \
            "Optional input list must be a list, not a {}.".format(type(optional_input_list))
        assert len(optional_inputs) == self.n_candidates, \
            "Length of the input list must be equal to number of candidates."
        out, mask = self.mutator.on_forward_input_choice(self, optional_input_list)
        if self.return_mask:
            return out, mask
        return out



#!/usr/bin/env python3

from autovul.base.configs import config, Config
from autovul.base.datasets import Dataset
from autovul.base.models import Model
from autovul.base.attacks import Attack
from autovul.base.utils import get_name
from autovul.base.utils.output import ansi
from autovul.base.utils.process import Model_Process

import os
from abc import ABC, abstractmethod

from typing import TYPE_CHECKING
from typing import Union    # TODO: python 3.10
import argparse    # TODO: python 3.10
if TYPE_CHECKING:
    pass


class Defense(ABC, Model_Process):
    name: str = 'defense'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        group.add_argument('--defense', dest='defense_name')
        group.add_argument('--defense_dir', help='directory to contain defense results')
        return group

    def __init__(self, attack: Attack = None, **kwargs):
        super().__init__(**kwargs)
        self.attack = attack

    @abstractmethod
    def detect(self, *args, **kwargs):
        ...


def add_argument(parser: argparse.ArgumentParser, defense_name: str = None, defense: Union[str, Defense] = None,
                 class_dict: dict[str, type[Defense]] = {}):
    defense_name = get_name(name=defense_name, module=defense, arg_list=['--defense'])
    group = parser.add_argument_group('{yellow}defense{reset}'.format(**ansi), description=defense_name)
    try:
        DefenseType = class_dict[defense_name]
    except KeyError as e:
        if defense_name is None:
            print('{red}you need to first claim the defense name using "--defense".{reset}'.format(**ansi))
        print(f'{defense_name} not in \n{list(class_dict.keys())}')
        raise e
    return DefenseType.add_argument(group)


def create(defense_name: str = None, defense: Union[str, Defense] = None, folder_path: str = None,
           dataset_name: str = None, dataset: Union[str, Dataset] = None,
           model_name: str = None, model: Union[str, Model] = None,
           config: Config = config, class_dict: dict[str, type[Defense]] = {}, **kwargs):
    dataset_name = get_name(name=dataset_name, module=dataset, arg_list=['-d', '--dataset'])
    model_name = get_name(name=model_name, module=model, arg_list=['-m', '--model'])
    defense_name = get_name(name=defense_name, module=defense, arg_list=['--defense'])
    if dataset_name is None:
        dataset_name = config.get_full_config()['dataset']['default_dataset']
    general_config = config.get_config(dataset_name=dataset_name)['defense']
    specific_config = config.get_config(dataset_name=dataset_name)[defense_name]
    result = general_config.update(specific_config).update(kwargs)    # TODO: linting issues
    try:
        DefenseType = class_dict[defense_name]
    except KeyError as e:
        print(f'{defense_name} not in \n{list(class_dict.keys())}')
        raise e
    if folder_path is None:
        folder_path = result['defense_dir']
        if isinstance(dataset, Dataset):
            folder_path = os.path.join(folder_path, dataset.data_type, dataset.name)
        if model_name is not None:
            folder_path = os.path.join(folder_path, model_name)
        folder_path = os.path.join(folder_path, DefenseType.name)
    return DefenseType(name=defense_name, dataset=dataset, model=model, folder_path=folder_path, **result)



#!/usr/bin/env python3

# CUDA_VISIBLE_DEVICES=0 python examples/validate.py --color --verbose 1 --dataset cifar10 --pretrain --model resnet18_comp

import autovul.vision
from autovul.vision.utils import summary
import argparse

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    autovul.vision.environ.add_argument(parser)
    autovul.vision.datasets.add_argument(parser)
    autovul.vision.models.add_argument(parser)
    args = parser.parse_args()

    env = autovul.vision.environ.create(**args.__dict__)
    dataset = autovul.vision.datasets.create(**args.__dict__)
    model = autovul.vision.models.create(dataset=dataset, **args.__dict__)



#!/usr/bin/env python3

from autovul.base import environ as environ
from autovul.base import datasets as datasets
from autovul.base import models as models
from autovul.base import trainer as trainer
from autovul.base.utils.tensor import to_tensor, to_numpy, to_list

__all__ = ['to_tensor', 'to_numpy', 'to_list']



#!/usr/bin/env python3

import autovul.base.optim

from autovul.base.utils import add_noise, cos_sim
from autovul.base.utils.output import prints
from autovul.base.environ import env

import torch
import torch.autograd
from collections.abc import Callable
from typing import Iterable, Union


def init_noise(noise_shape: Iterable[int], pgd_eps: Union[float, torch.Tensor],
               random_init: bool = False, device: Union[str, torch.device] = None) -> torch.Tensor:
    device = device if device is not None else env['device']
    noise: torch.Tensor = torch.zeros(noise_shape, dtype=torch.float, device=device)
    if random_init:
        if isinstance(pgd_eps, torch.Tensor) and pgd_eps.shape[0] != 1:
            assert all([size == 1 for size in pgd_eps.shape[1:]])
            for i in range(pgd_eps.shape[0]):
                data = noise[i, :, :] if noise.dim() == 3 else noise[:, i, :, :]
                data.uniform_(-pgd_eps[i].item(), pgd_eps[i].item())
        else:
            pgd_eps = float(pgd_eps)
            noise.uniform_(-pgd_eps, pgd_eps)
    return noise


def valid_noise(adv_input: torch.Tensor, org_input: torch.Tensor, universal: bool = False) -> torch.Tensor:
    result = (adv_input - org_input).detach()
    return result.mode(dim=0)[0] if universal else result


class PGDoptimizer(autovul.base.optim.Optimizer):
    r"""Projected Gradient Descent.
    Args:
        pgd_alpha (float): learning rate :math:`\pgd_alpha`. Default: :math:`\frac{3}{255}`.
        pgd_eps (float): the perturbation threshold :math:`\pgd_eps` in input space. Default: :math:`\frac{8}{255}`.

        norm (int): :math:`L_p` norm passed to :func:`torch.norm`. Default: ``float(inf)``.
        universal (bool): All inputs in the batch share the same noise. Default: ``False``.

        grad_method (str): gradient estimation method (['white', 'nes', 'sgd', 'hess', 'zoo']). Default: ``white``.
        query_num (int): number of samples in black box gradient estimation. Default: ``100``.
        sigma (float): gaussian noise std in black box gradient estimation. Default: ``0.001``.
    """

    name: str = 'pgd'

    def __init__(self, pgd_alpha: Union[float, torch.Tensor] = 2.0 / 255,
                 pgd_eps: Union[float, torch.Tensor] = 8.0 / 255,
                 iteration: int = 7, random_init: bool = False,
                 norm: Union[int, float] = float('inf'), universal: bool = False,
                 clip_min: Union[float, torch.Tensor] = 0.0,
                 clip_max: Union[float, torch.Tensor] = 1.0,
                 grad_method: str = 'white', query_num: int = 100, sigma: float = 1e-3,
                 hess_b: int = 100, hess_p: int = 1, hess_lambda: float = 1, **kwargs):
        super().__init__(iteration=iteration, **kwargs)
        self.param_list['pgd'] = ['pgd_alpha', 'pgd_eps', 'random_init', 'norm', 'universal']

        pgd_alpha = pgd_alpha if pgd_alpha < 1.0 else pgd_alpha / 255
        pgd_eps = pgd_eps if pgd_eps < 1.0 else pgd_eps / 255

        self.pgd_alpha = pgd_alpha
        self.pgd_eps = pgd_eps
        self.random_init = random_init

        self.norm = norm
        self.universal = universal

        self.clip_min = clip_min
        self.clip_max = clip_max

        self.grad_method: str = grad_method
        if grad_method != 'white':
            self.param_list['blackbox'] = ['grad_method', 'query_num', 'sigma']
            self.query_num: int = query_num
            self.sigma: float = sigma
            if grad_method == 'hess':
                self.param_list['hessian'] = ['hess_b', 'hess_p', 'hess_lambda']
                self.hess_b: int = hess_b
                self.hess_p: int = hess_p
                self.hess_lambda: float = hess_lambda

    def optimize(self, _input: torch.Tensor,
                 noise: torch.Tensor = None,
                 pgd_alpha: Union[float, torch.Tensor] = None,
                 pgd_eps: Union[float, torch.Tensor] = None,
                 add_noise_fn: Callable[..., torch.Tensor] = None,
                 random_init: bool = None,
                 clip_min: Union[float, torch.Tensor] = None,
                 clip_max: Union[float, torch.Tensor] = None,
                 *args, **kwargs) -> tuple[torch.Tensor, torch.Tensor]:
        # ------------------------------ Parameter Initialization ---------------------------------- #
        clip_min = clip_min if clip_min is not None else self.clip_min
        clip_max = clip_max if clip_max is not None else self.clip_max
        pgd_alpha = pgd_alpha if pgd_alpha is not None else self.pgd_alpha
        pgd_eps = pgd_eps if pgd_eps is not None else self.pgd_eps
        random_init = random_init if random_init is not None else self.random_init
        add_noise_fn = add_noise_fn if add_noise_fn is not None else add_noise
        if noise is None:
            noise_shape = _input.shape[1:] if self.universal else _input.shape
            noise = self.init_noise(noise_shape, pgd_eps=pgd_eps, random_init=random_init, device=_input.device)
        # ----------------------------------------------------------------------------------------- #
        a = pgd_alpha if isinstance(pgd_alpha, torch.Tensor) else torch.tensor(pgd_alpha)
        b = pgd_eps if isinstance(pgd_eps, torch.Tensor) else torch.tensor(pgd_eps)
        condition_alpha = a.allclose(torch.zeros_like(a))
        condition_eps = b.allclose(torch.zeros_like(b))
        if condition_alpha or condition_eps:
            return _input, None
        # ----------------------------------------------------------------------------------------- #
        kwargs.update(noise=noise,
                      pgd_alpha=pgd_alpha, pgd_eps=pgd_eps,
                      add_noise_fn=add_noise_fn, random_init=random_init,
                      clip_min=clip_min, clip_max=clip_max)
        return super().optimize(_input, *args, **kwargs)

    def update_input(self, current_idx: torch.Tensor,
                     adv_input: torch.Tensor,
                     org_input: torch.Tensor,
                     noise: torch.Tensor,
                     pgd_alpha: Union[float, torch.Tensor],
                     pgd_eps: Union[float, torch.Tensor],
                     add_noise_fn: Callable[..., torch.Tensor],
                     clip_min: Union[float, torch.Tensor],
                     clip_max: Union[float, torch.Tensor],
                     loss_fn: Callable[[torch.Tensor], torch.Tensor],
                     output: list[str],
                     loss_kwargs: dict[str, torch.Tensor] = {},
                     *args, **kwargs):
        current_loss_kwargs = {k: v[current_idx] for k, v in loss_kwargs.items()}
        grad = self.calc_grad(loss_fn, adv_input[current_idx], loss_kwargs=current_loss_kwargs)
        if self.grad_method != 'white' and 'middle' in output:
            real_grad = self.whitebox_grad(loss_fn, adv_input[current_idx], loss_kwargs=current_loss_kwargs)
            prints('cos<real, est> = ', cos_sim(grad.sign(), real_grad.sign()),
                   indent=self.indent + 2)
        if self.universal:
            grad = grad.mean(dim=0)
        noise[current_idx] = (noise[current_idx] - pgd_alpha * torch.sign(grad))
        noise[current_idx] = self.projector(noise[current_idx], pgd_eps, norm=self.norm)
        adv_input[current_idx] = add_noise_fn(_input=org_input[current_idx], noise=noise[current_idx], batch=self.universal,
                                              clip_min=clip_min, clip_max=clip_max)
        noise[current_idx] = self.valid_noise(adv_input[current_idx], org_input[current_idx])

    def preprocess_input(self, adv_input: torch.Tensor, org_input: torch.Tensor,
                         noise: torch.Tensor = None,
                         add_noise_fn: Callable[..., torch.Tensor] = None,
                         clip_min: Union[float, torch.Tensor] = None,
                         clip_max: Union[float, torch.Tensor] = None,
                         *args, **kwargs) -> torch.Tensor:
        adv_input = add_noise_fn(_input=adv_input, noise=noise, batch=self.universal,
                                 clip_min=clip_min, clip_max=clip_max)
        noise.copy_(self.valid_noise(adv_input, org_input))
        return adv_input

    def output_info(self, org_input: torch.Tensor, noise: torch.Tensor,
                    loss_fn: Callable[[torch.Tensor], torch.Tensor] = None,
                    loss_kwargs: dict[str, torch.Tensor] = {},
                    *args, **kwargs):
        super().output_info(*args, **kwargs)
        with torch.no_grad():
            loss = float(loss_fn(org_input + noise, **loss_kwargs))
            norm = noise.norm(p=self.norm)
            prints(f'L-{self.norm} norm: {norm}    loss: {loss:.5f}', indent=self.indent)

    def valid_noise(self, adv_input: torch.Tensor, org_input: torch.Tensor, universal: bool = None) -> torch.Tensor:
        universal = universal if universal is not None else self.universal
        return valid_noise(adv_input, org_input, universal=universal)

    def init_noise(self, noise_shape: Iterable[int], pgd_eps: Union[float, torch.Tensor] = None,
                   random_init: bool = None, device: Union[str, torch.device] = None) -> torch.Tensor:
        pgd_eps = pgd_eps if pgd_eps is not None else self.pgd_eps
        random_init = random_init if random_init is not None else self.random_init
        return init_noise(noise_shape, pgd_eps, random_init=random_init, device=device)

    @staticmethod
    def projector(noise: torch.Tensor, pgd_eps: Union[float, torch.Tensor],
                  norm: Union[float, int, str] = float('inf')) -> torch.Tensor:
        if norm == float('inf'):
            noise = noise.clamp(min=-pgd_eps, max=pgd_eps)
        elif isinstance(pgd_eps, float):
            norm: torch.Tensor = noise.flatten(-3).norm(p=norm, dim=-1)
            length = pgd_eps / norm.unsqueeze(-1).unsqueeze(-1)
            noise = length * noise
        else:
            norm = noise.flatten(-2).norm(p=norm, dim=-1)
            length = pgd_eps / norm.unsqueeze(-1).unsqueeze(-1)
            noise = length * noise
        return noise.detach()

    # -------------------------- Calculate Gradient ------------------------ #
    def calc_grad(self, f, x: torch.Tensor, grad_method: str = None, loss_kwargs: dict[str, torch.Tensor] = {}) -> torch.Tensor:
        grad_method = grad_method if grad_method is not None else self.grad_method
        grad_func = self.whitebox_grad if grad_method == 'white' else self.blackbox_grad
        return grad_func(f, x, loss_kwargs=loss_kwargs)

    @staticmethod
    def whitebox_grad(f, x: torch.Tensor, loss_kwargs: dict[str, torch.Tensor] = {}) -> torch.Tensor:
        x.requires_grad_()
        loss = f(x, **loss_kwargs)
        grad = torch.autograd.grad(loss, x)[0]
        x.requires_grad_(False)
        return grad

    def blackbox_grad(self, f: Callable[[torch.Tensor], torch.Tensor], x: torch.Tensor,
                      query_num: int = None, sigma: float = None, loss_kwargs: dict[str, torch.Tensor] = {}) -> torch.Tensor:
        seq = self.gen_seq(x, query_num=query_num, sigma=sigma)
        grad = self.calc_seq(f, seq, loss_kwargs=loss_kwargs)
        return grad

    # x: (N, C, H, W)
    # return: (query_num+1, N, C, H, W)
    def gen_seq(self, x: torch.Tensor, query_num: int = None, sigma: float = None) -> torch.Tensor:
        query_num = query_num if query_num is not None else self.query_num
        sigma = sigma if sigma is not None else self.sigma
        shape = list(x.shape)
        shape.insert(0, query_num)
        if self.grad_method == 'nes':
            shape[0] = shape[0] // 2
        noise = sigma * torch.normal(mean=0.0, std=1.0, size=shape, device=x.device)

        zeros = torch.zeros_like(x.unsqueeze(0))
        seq = [zeros]
        if self.grad_method == 'nes':
            seq.extend([noise, -noise])
            if query_num % 2 == 1:
                seq.append(zeros)
        elif self.grad_method == 'sgd':
            seq.append(noise)
        elif self.grad_method == 'hess':
            raise NotImplementedError(self.grad_method)
            noise = (self.hess @ noise.view(-1, 1)).view(X.shape)
            seq.append(noise)
        elif self.grad_method == 'zoo':
            raise NotImplementedError(self.grad_method)
        else:
            print('Current method: ', self.grad_method)
            raise ValueError("Argument 'method' should be 'nes', 'sgd' or 'hess'!")
        seq = torch.cat(seq).add(x)  # (query_num+1, N, C, H, W)
        return seq

    def calc_seq(self, f: Callable[[torch.Tensor], torch.Tensor], seq: torch.Tensor,
                 loss_kwargs: dict[str, torch.Tensor] = {}) -> torch.Tensor:
        X = seq[0]  # (N, C, H, W)
        seq = seq[1:]  # (query_num, N, C, H, W)
        noise = seq.sub(X)
        with torch.no_grad():
            temp_list: list[torch.Tensor] = []
            for sub_seq in seq:
                temp_list.append(f(sub_seq, reduction='none', **loss_kwargs))   # (query_num, N)
            g = torch.stack(temp_list)[:, :, None, None, None].mul(noise).sum(dim=0)  # (N, C, H, W)
            if self.grad_method in ['sgd', 'hess']:
                g -= f(X) * noise.sum(dim=0)
            g /= len(seq) * self.sigma * self.sigma
        return g

    @staticmethod
    def calc_hess(f: Callable[[torch.Tensor], torch.Tensor], X: torch.Tensor,
                  sigma: float, hess_b: int, hess_lambda: float = 1) -> torch.Tensor:
        length = X.numel()
        hess: torch.Tensor = torch.zeros(length, length, device=X.device)
        with torch.no_grad():
            for i in range(hess_b):
                noise = torch.normal(mean=0.0, std=1.0, size=X.shape, device=X.device)
                X1 = X + sigma * noise
                X2 = X - sigma * noise
                hess += abs(f(X1) + f(X2) - 2 * f(X)) * \
                    (noise.view(-1, 1) @ noise.view(1, -1))
            hess /= (2 * hess_b * sigma * sigma)
            hess += hess_lambda * torch.eye(length, device=X.device)
            result = hess.cholesky_inverse()
        return result



#!/usr/bin/env python3


import torch
import torch.nn as nn
import torch.nn.functional as F

from collections import OrderedDict


def get_op(op_name: str, C_in: int, stride: int = 1, affine: bool = True, p: float = None,
           C_out: int = None, std_conv: bool = False):
    C_out = C_out if C_out is not None else C_in
    if op_name == 'none':
        return Zero(stride)
    elif op_name == 'noise':
        return Noise(stride)
    elif op_name == 'skip_connect' and stride == 1:
        return nn.Identity()
    else:
        seq = nn.Sequential()
        if 'pool' not in op_name and 'sep_conv' not in op_name and 'dil_conv' not in op_name:
            seq.add_module('relu', nn.ReLU())

        if op_name == 'conv':
            ConvClass = StdConv2d if std_conv else nn.Conv2d
            seq.add_module('conv', ConvClass(C_in, C_out, 1, stride, 0, bias=False))
        elif op_name == 'avg_pool_3x3':
            seq.add_module('pool', nn.AvgPool2d(3, stride, 1, count_include_pad=False))
        elif op_name == 'max_pool_2x2':
            seq.add_module('pool', nn.MaxPool2d(2, stride, 0))
        elif op_name == 'max_pool_3x3':
            seq.add_module('pool', nn.MaxPool2d(3, stride, 1))
        elif op_name == 'max_pool_5x5':
            seq.add_module('pool', nn.MaxPool2d(5, stride, 2))
        elif op_name in ['skip_connect', 'factorized_reduce']:
            seq.add_module('reduce', FactorizedReduce(C_in, C_out))
        elif op_name == 'sep_conv_3x3':
            seq.add_module('dil_conv1', DilConv(C_in, C_out, 3, stride, 1, dilation=1,
                                                std_conv=std_conv, affine=affine))
            seq.add_module('dil_conv2', DilConv(C_in, C_out, 3, 1, 1, dilation=1,
                                                std_conv=std_conv, affine=affine))
        elif op_name == 'sep_conv_5x5':
            seq.add_module('dil_conv1', DilConv(C_in, C_out, 5, stride, 2, dilation=1,
                                                std_conv=std_conv, affine=affine))
            seq.add_module('dil_conv2', DilConv(C_in, C_out, 5, 1, 2, dilation=1,
                                                std_conv=std_conv, affine=affine))
        elif op_name == 'sep_conv_7x7':
            seq.add_module('dil_conv1', DilConv(C_in, C_out, 7, stride, 3, dilation=1,
                                                std_conv=std_conv, affine=affine))
            seq.add_module('dil_conv2', DilConv(C_in, C_out, 7, 1, 3, dilation=1,
                                                std_conv=std_conv, affine=affine))
        elif op_name == 'dil_conv_3x3':
            seq = DilConv(C_in, C_out, 3, stride, 2, dilation=2,
                          std_conv=std_conv, affine=affine)
        elif op_name == 'dil_conv_5x5':
            seq = DilConv(C_in, C_out, 5, stride, 4, dilation=2,
                          std_conv=std_conv, affine=affine)
        elif op_name == 'conv_7x1_1x7':
            seq.add_module('conv1', nn.Conv2d(C_in, C_out, (1, 7), (1, stride), (0, 3), bias=False))
            seq.add_module('conv2', nn.Conv2d(C_in, C_out, (7, 1), (stride, 1), (3, 0), bias=False))
        elif op_name == 'conv_1x1':
            seq = nn.Conv2d(C_in, C_out, 1, stride, 0, bias=False)
        elif op_name == 'conv_3x3':
            seq = nn.Conv2d(C_in, C_out, 3, stride, 1, bias=False)
        elif op_name == 'conv_5x5':
            seq = nn.Conv2d(C_in, C_out, 5, stride, 2, bias=False)

        if 'pool' not in op_name and 'sep_conv' not in op_name and 'dil_conv' not in op_name:
            seq.add_module('bn', nn.BatchNorm2d(C_out, affine=affine))
        if p is not None:
            seq.add_module('dropout', nn.Dropout(p=p))
        return seq


def DilConv(C_in: int, C_out: int, kernel_size: int, stride: int, padding: int, dilation: int,
            std_conv: bool = False, affine: bool = True):
    """ (Dilated) depthwise separable conv
    ReLU - (Dilated) depthwise separable - Pointwise - BN
    If dilation == 2, 3x3 conv => 5x5 receptive field
                      5x5 conv => 9x9 receptive field
    """
    ConvClass = StdConv2d if std_conv else nn.Conv2d
    return nn.Sequential(OrderedDict([
        ('relu', nn.ReLU()),
        ('conv1', nn.Conv2d(C_in, C_in, kernel_size, stride, padding, dilation=dilation, groups=C_in, bias=False)),
        ('conv2', ConvClass(C_in, C_out, 1, stride=1, padding=0, bias=False)),
        ('bn', nn.BatchNorm2d(C_out, affine=affine))
    ]))


class Zero(nn.Module):
    def __init__(self, stride: int = 1):
        super().__init__()
        self.stride = stride

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return x.mul(0.) if self.stride == 1 else x[:, :, ::self.stride, ::self.stride].mul(0.)


class Noise(nn.Module):
    def __init__(self, stride: int = 1, mean: float = 0.0, std: float = 1.0):
        super().__init__()
        self.stride = stride
        self.mean = mean
        self.std = std

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        result = torch.randn_like(x).mul_(self.std).add_(self.std)
        return result if self.stride == 1 else result[:, :, ::self.stride, ::self.stride]


class FactorizedReduce(nn.Module):
    """ Reduce feature map size by factorized pointwise(stride=2). """

    def __init__(self, C_in: int, C_out: int):
        super().__init__()
        C_out_1 = C_out // 2
        self.conv1 = nn.Conv2d(C_in, C_out_1, 1, stride=2, padding=0, bias=False)
        self.conv2 = nn.Conv2d(C_in, C_out - C_out_1, 1, stride=2, padding=0, bias=False)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return torch.cat([self.conv1(x), self.conv2(x[:, :, 1:, 1:])], dim=1)


class StdConv2d(nn.Conv2d):
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        weight = self.weight.flatten(start_dim=1)
        weight = weight.sub(weight.mean(dim=1, keepdim=True)).div(
            weight.std(dim=1, keepdim=True) + 1e-5).view_as(self.weight)
        return F.conv2d(x, weight, self.bias, self.stride,
                        self.padding, self.dilation, self.groups)



#!/usr/bin/env python3

# https://github.com/rwightman/pytorch-dpn-pretrained
# https://github.com/rwightman/pytorch-image-models/tree/master/timm/models
# https://github.com/kuangliu/pytorch-cifar/blob/master/models/dpn.py

'''Dual Path Networks in PyTorch.'''
import torch
import torch.nn as nn

from torchvision.models.resnet import conv1x1, conv3x3

from collections.abc import Callable
from typing import Optional


class Bottleneck(nn.Module):
    def __init__(self, inplanes: int, planes: int,
                 out_planes: int, dense_depth: int,
                 stride: int = 1, downsample: nn.Module = None,
                 groups: int = 1,
                 base_width: int = 64,
                 dilation: int = 1,
                 norm_layer: Optional[Callable[..., nn.Module]] = None) -> None:
        super().__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        width = int(planes * (base_width / 64.)) * groups
        # Both self.conv2 and self.downsample layers downsample the input when stride != 1
        self.conv1 = conv1x1(inplanes, width)
        self.bn1 = norm_layer(width)
        self.conv2 = conv3x3(width, width, stride, groups, dilation)
        self.bn2 = norm_layer(width)
        self.conv3 = conv1x1(width, out_planes + dense_depth)   # difference
        self.bn3 = norm_layer(out_planes + dense_depth)
        self.downsample = downsample
        self.relu = nn.ReLU(inplace=True)
        self.stride = stride
        self.out_planes = out_planes    # difference
        self.dense_depth = dense_depth

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        identity = x

        out = self.conv1(x)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)
        out = self.relu(out)

        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        d = self.out_planes  # difference
        out = torch.cat([identity[:, :d, :, :] + out[:, :d, :, :],
                         identity[:, d:, :, :], out[:, d:, :, :]], dim=1)
        out = self.relu(out)

        return out


class DPN(nn.Module):
    def __init__(self, block: type[Bottleneck],
                 num_init_features: int,
                 planes: list[int],
                 layers: list[int],
                 out_planes: list[int],
                 dense_depth: list[int],
                 num_classes: int = 1000,
                 zero_init_residual: bool = False,
                 groups: int = 1,
                 width_per_group: int = 64,
                 replace_stride_with_dilation: Optional[list[bool]] = None,
                 norm_layer: Optional[Callable[..., nn.Module]] = None,
                 small: bool = False) -> None:
        super().__init__()
        if norm_layer is None:
            norm_layer = nn.BatchNorm2d
        self._norm_layer = norm_layer

        self.inplanes = num_init_features
        self.dilation = 1
        if replace_stride_with_dilation is None:
            # each element in the tuple indicates if we should replace
            # the 2x2 stride with a dilated convolution instead
            replace_stride_with_dilation = [False, False, False]
        if len(replace_stride_with_dilation) != 3:
            raise ValueError("replace_stride_with_dilation should be None "
                             "or a 3-element tuple, got {}".format(replace_stride_with_dilation))
        self.groups = groups
        self.base_width = width_per_group
        self.conv1 = nn.Conv2d(3, self.inplanes, kernel_size=3 if small else 7, stride=2,
                               padding=1 if small else 3, bias=False)
        self.bn1 = norm_layer(self.inplanes)
        self.relu = nn.ReLU(inplace=True)
        self.maxpool = nn.MaxPool2d(kernel_size=3, stride=2, padding=1)
        self.layer1 = self._make_layer(block, planes[0], layers[0], out_planes[0], dense_depth[0])
        self.layer2 = self._make_layer(block, planes[1], layers[1], out_planes[1], dense_depth[1], stride=2,
                                       dilate=replace_stride_with_dilation[0])
        self.layer3 = self._make_layer(block, planes[2], layers[2], out_planes[2], dense_depth[2], stride=2,
                                       dilate=replace_stride_with_dilation[1])
        self.layer4 = self._make_layer(block, planes[3], layers[3], out_planes[3], dense_depth[3], stride=2,
                                       dilate=replace_stride_with_dilation[2])
        self.avgpool = nn.AdaptiveAvgPool2d((1, 1))
        self.fc = nn.Linear(out_planes[3] + (layers[3] + 1) * dense_depth[3], num_classes)

        for m in self.modules():
            if isinstance(m, nn.Conv2d):
                nn.init.kaiming_normal_(m.weight, mode='fan_out', nonlinearity='relu')
            elif isinstance(m, (nn.BatchNorm2d, nn.GroupNorm)):
                nn.init.constant_(m.weight, 1)
                nn.init.constant_(m.bias, 0)

        # Zero-initialize the last BN in each residual branch,
        # so that the residual branch starts with zeros, and each residual block behaves like an identity.
        # This improves the model by 0.2~0.3% according to https://arxiv.org/abs/1706.02677
        if zero_init_residual:
            for m in self.modules():
                if isinstance(m, Bottleneck):
                    nn.init.constant_(m.bn3.weight, 0)  # type: ignore[arg-type]
                # elif isinstance(m, BasicBlock):
                #     nn.init.constant_(m.bn2.weight, 0)  # type: ignore[arg-type]

    def _make_layer(self, block: type[Bottleneck], planes: int, blocks: int,
                    out_planes: int, dense_depth: int,
                    stride: int = 1, dilate: bool = False) -> nn.Sequential:
        norm_layer = self._norm_layer
        # downsample = None
        previous_dilation = self.dilation
        if dilate:
            self.dilation *= stride
            stride = 1
        # if stride != 1 or self.inplanes != planes * block.expansion:
        downsample = nn.Sequential(
            conv1x1(self.inplanes, out_planes + dense_depth, stride),
            norm_layer(out_planes + dense_depth),
        )

        layers = []
        layers.append(block(self.inplanes, planes, out_planes, dense_depth,
                            stride, downsample, self.groups,
                            self.base_width, previous_dilation, norm_layer))
        self.inplanes = out_planes + 2 * dense_depth
        for i in range(1, blocks):
            layers.append(block(self.inplanes, planes, out_planes, dense_depth, groups=self.groups,
                                base_width=self.base_width, dilation=self.dilation,
                                norm_layer=norm_layer))
            self.inplanes = out_planes + (2 + i) * dense_depth

        return nn.Sequential(*layers)

    def _forward_impl(self, x: torch.Tensor) -> torch.Tensor:
        # See note [TorchScript super()]
        x = self.conv1(x)
        x = self.bn1(x)
        x = self.relu(x)
        x = self.maxpool(x)

        x = self.layer1(x)
        x = self.layer2(x)
        x = self.layer3(x)
        x = self.layer4(x)

        x = self.avgpool(x)
        x = torch.flatten(x, 1)
        x = self.fc(x)

        return x

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self._forward_impl(x)


def dpn68(**kwargs):
    return DPN(Bottleneck, num_init_features=10,
               planes=[4, 8, 16, 32], layers=[3, 4, 12, 3],
               out_planes=[64, 128, 256, 512], dense_depth=[16, 32, 32, 64],
               groups=32, small=True, **kwargs)


def dpn92(**kwargs):
    return DPN(Bottleneck, num_init_features=64,
               planes=[3, 6, 12, 24], layers=[3, 4, 20, 3],
               out_planes=[256, 512, 1024, 2048], dense_depth=[16, 32, 24, 128],
               groups=32, **kwargs)


def dpn98(**kwargs):
    return DPN(Bottleneck, num_init_features=96,
               planes=[4, 8, 16, 32], layers=[3, 6, 20, 3],
               out_planes=[256, 512, 1024, 2048], dense_depth=[16, 32, 32, 128],
               groups=40, **kwargs)


def dpn131(**kwargs):
    return DPN(Bottleneck, num_init_features=128,
               planes=[4, 8, 16, 32], layers=[4, 8, 28, 3],
               out_planes=[256, 512, 1024, 2048], dense_depth=[16, 32, 32, 128],
               groups=40, **kwargs)


def dpn107(**kwargs):
    return DPN(Bottleneck, num_init_features=128,
               planes=[5, 10, 20, 40], layers=[4, 8, 20, 3],
               out_planes=[256, 512, 1024, 2048], dense_depth=[20, 64, 64, 128],
               groups=40, **kwargs)



#!/usr/bin/env python3

# https://github.com/Thrandis/EKFAC-pytorch/blob/master/kfac.py
# https://github.com/n-gao/pytorch-kfac

from autovul.base.utils.lock import Lock

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.optim.optimizer import Optimizer

import math
from abc import ABC, abstractmethod

from typing import Iterable, Union, Optional
from torch.utils.hooks import RemovableHandle

LayerType = Union[nn.Conv2d, nn.Linear]
_LayerType = (nn.Conv2d, nn.Linear)


class BaseState:
    def __init__(self):
        self.x: torch.Tensor = None   # (N, in, xh, xw)
        self.gy: torch.Tensor = None   # (N, out, yh, yw)
        self.num_locations: int = None  # yh * yw


class KFACState(BaseState):
    def __init__(self):
        super().__init__()
        self.xxt: torch.Tensor = None   # (in [* kh * kw] + 1, in [* kh * kw] + 1)
        self.ggt: torch.Tensor = None   # (out, out)
        self.ixxt: torch.Tensor = None   # (in [* kh * kw] + 1, in [* kh * kw] + 1)
        self.iggt: torch.Tensor = None   # (out, out)


class BaseKFAC(ABC, Optimizer):
    def __init__(self, net: nn.Module, eps: float = 0.1,
                 sua: bool = False, update_freq: int = 1,
                 alpha: float = 1.0, constraint_norm: bool = False,
                 state_type: type = BaseState):
        """ Base K-FAC Preconditionner for Linear and Conv2d layers.

        Computes the K-FAC of the second moment of the gradients.
        It works for Linear and Conv2d layers and silently skip other layers.

        Args:
            net (torch.nn.Module): Network to precondition.
            eps (float): Tikhonov regularization parameter for the inverses.
            sua (bool): Applies SUA approximation.
            update_freq (int): Perform inverses every update_freq updates.
            alpha (float): Running average parameter (if == 1, no r. ave.).
            constraint_norm (bool): Scale the gradients by the squared
                fisher norm.
        """
        self.track = Lock()

        self.eps = eps
        self.sua = sua
        self.update_freq = update_freq
        self.alpha = alpha
        self.constraint_norm = constraint_norm
        self._fwd_handles: list[RemovableHandle] = []
        self._bwd_handles: list[RemovableHandle] = []
        self._iteration_counter = 0
        params = []
        # [mod for mod in net.modules() if isinstance(mod, _LayerType)]
        name_dict: dict[str, LayerType] = {}
        for name, mod in net.named_modules():
            if isinstance(mod, _LayerType):
                handle = mod.register_forward_pre_hook(self._save_input)
                self._fwd_handles.append(handle)
                handle = mod.register_full_backward_hook(self._save_grad_output)
                self._bwd_handles.append(handle)
                params.append({'params': mod.parameters(), 'mod': mod})
                if len(name):
                    name += '.'
                name_dict[name + 'weight'] = mod
                if mod.bias is not None:
                    name_dict[name + 'bias'] = mod
        super().__init__(params, {})
        self.param_groups: list[dict[str, Union[LayerType, Iterable[torch.Tensor]]]]
        self.module_list: list[LayerType] = [group['mod'] for group in self.param_groups]
        self.pack_idx: dict[LayerType, list[int]] = {module: [] for module in self.module_list}
        for i, (name, _) in enumerate(net.named_parameters()):
            if name in name_dict.keys():
                self.pack_idx[name_dict[name]].append(i)
        self.state_storage: dict[LayerType, BaseState] = {mod: state_type() for mod in self.module_list}

    def step(self, update_stats: bool = True, update_params: bool = True, force: bool = False):
        """Performs one step of preconditioning."""
        if update_stats and (force or self._iteration_counter % self.update_freq == 0):
            self.update_stats()
        if update_params:
            self.update_params()
        if update_stats:
            self._iteration_counter += 1

    @abstractmethod
    def update_stats(self, **kwargs):
        ...

    def update_params(self, constraint_norm: bool = None):
        if constraint_norm is None:
            constraint_norm = self.constraint_norm
        fisher_norm = 0.

        for module in self.module_list:
            weight_grad = module.weight.grad
            bias_grad = None if module.bias is None else module.bias.grad
            # Preconditionning
            gw, gb = self.precond(module, weight_grad, bias_grad)
            # Updating gradients
            if constraint_norm:
                fisher_norm += (weight_grad * gw).sum()
            weight_grad.data = gw
            if bias_grad is not None:
                if constraint_norm:
                    fisher_norm += (bias_grad * gb).sum()
                bias_grad.data = gb
        # Eventually scale the norm of the gradients
        if constraint_norm:
            scale = math.sqrt(1. / fisher_norm)
            for module in self.module_list:
                for param in module.parameters():
                    param.grad.data *= scale

    def calc_grad(self, grad_list: list[torch.Tensor], constraint_norm: bool = None) -> list[torch.Tensor]:
        if constraint_norm is None:
            constraint_norm = self.constraint_norm
        if grad_list is None:
            grad_dict = {module: [param.grad for param in module.parameters()]
                         for module in self.module_list}
        else:
            if not isinstance(grad_list, list):
                grad_list = list(grad_list)
            grad_dict = self.pack_list(grad_list, self.pack_idx)
        fisher_norm = 0.

        for module, grads in grad_dict.items():
            weight_grad = grads[0]
            bias_grad = None if len(grads) == 0 else grads[1]
            # Preconditionning
            gw, gb = self.precond(module, weight_grad, bias_grad)
            # Updating gradients
            if constraint_norm:
                fisher_norm += (weight_grad * gw).sum()
            grads[0] = gw
            if bias_grad is not None:
                if constraint_norm:
                    fisher_norm += (bias_grad * gb).sum()
                grads[1] = gb
        # Eventually scale the norm of the gradients
        if constraint_norm:
            scale = math.sqrt(1. / fisher_norm)
            for grads in grad_dict.values():
                for i in range(len(grads)):
                    grads[i] *= scale
        return self.unpack_list(grad_list, grad_dict, self.pack_idx)

    def precond(self, mod: LayerType, weight_grad: torch.Tensor = None,
                bias_grad: Optional[torch.Tensor] = None
                ) -> tuple[torch.Tensor, Optional[torch.Tensor]]:
        """Applies preconditioning."""
        if weight_grad is None and bias_grad is None:
            weight_grad = mod.weight.grad
            if mod.bias is not None:
                bias_grad = mod.bias.grad
        precond_func = self.precond_sua if isinstance(mod, nn.Conv2d) and self.sua else self.precond_nosua
        return precond_func(mod, weight_grad, bias_grad)

    @abstractmethod
    def precond_sua(self, mod: nn.Conv2d, weight_grad: torch.Tensor,
                    bias_grad: Optional[torch.Tensor]
                    ) -> tuple[torch.Tensor, Optional[torch.Tensor]]:
        ...

    @abstractmethod
    def precond_nosua(self, mod: nn.Conv2d, weight_grad: torch.Tensor,
                      bias_grad: Optional[torch.Tensor]
                      ) -> tuple[torch.Tensor, Optional[torch.Tensor]]:
        ...

    @torch.no_grad()
    def _save_input(self, mod: LayerType, i: tuple[torch.Tensor]):
        """Saves input of layer to compute covariance."""
        if self.track and mod.training:
            self.state_storage[mod].x = i[0].detach().clone()

    @torch.no_grad()
    def _save_grad_output(self, mod: LayerType,
                          grad_input: tuple[torch.Tensor],
                          grad_output: tuple[torch.Tensor]):
        """Saves grad on output of layer to compute covariance."""
        if self.track and mod.training:
            gy = grad_output[0]
            self.state_storage[mod].gy = gy.size(0) * gy.detach().clone()

    def reset(self):
        for k, v in self.state_storage.items():
            self.state_storage[k] = type(v)()
        self._iteration_counter = 0

    def __del__(self):
        for handle in self._fwd_handles + self._bwd_handles:
            handle.remove()

    @staticmethod
    def pack_list(param_list: list[torch.Tensor], pack_idx: dict[nn.Module, list[int]]) -> dict[nn.Module, list[torch.Tensor]]:
        return {module: [param_list[idx] for idx in idx_list] for module, idx_list in pack_idx.items()}

    @staticmethod
    def unpack_list(param_list: list[torch.Tensor], param_dict: dict[nn.Module, list[torch.Tensor]],
                    pack_idx: dict[nn.Module, list[int]]) -> dict[nn.Module, list[torch.Tensor]]:
        new_list = [param for param in param_list]
        for module, idx_list in pack_idx.items():
            for i, idx in enumerate(idx_list):
                new_list[idx] = param_dict[module][i]
        return new_list


class KFAC(BaseKFAC):

    def __init__(self, *args, pi: bool = False, **kwargs):
        """ K-FAC Preconditionner for Linear and Conv2d layers.

        Computes the K-FAC of the second moment of the gradients.
        It works for Linear and Conv2d layers and silently skip other layers.

        Args:
            net (torch.nn.Module): Network to precondition.
            pi (bool): Computes pi correction for Tikhonov regularization.
            eps (float): Tikhonov regularization parameter for the inverses.
            sua (bool): Applies SUA approximation.
            update_freq (int): Perform inverses every update_freq updates.
            alpha (float): Running average parameter (if == 1, no r. ave.).
            constraint_norm (bool): Scale the gradients by the squared
                fisher norm.
        """
        super().__init__(*args, state_type=KFACState, **kwargs)
        self.pi = pi
        self.state_storage: dict[LayerType, KFACState]

    def update_stats(self, **kwargs):
        for module in self.module_list:
            self.compute_covs(module)
            state = self.state_storage[module]
            state.ixxt, state.iggt = self.inv_covs(state.xxt, state.ggt,
                                                   num_locations=state.num_locations,
                                                   eps=self.eps, pi=self.pi)

    def precond_sua(self, mod: nn.Conv2d, weight_grad: torch.Tensor,
                    bias_grad: Optional[torch.Tensor]
                    ) -> tuple[torch.Tensor, Optional[torch.Tensor]]:
        """Preconditioning for KFAC SUA."""
        state = self.state_storage[mod]
        g, gb = weight_grad, bias_grad
        s = weight_grad.size()  # (out, in, kh, kw)
        if bias_grad is not None:
            gb = gb[None, :, None, None].repeat(1, 1, s[2], s[3])  # (out, 1, kh, kw)
            g = torch.cat([g, gb], dim=1)  # (out, in + 1, kh, kw)

        g = g.transpose(0, 1)  # (in + 1, out, kh, kw)
        g = state.ixxt.mm(g.flatten(1)).view_as(g).transpose(0, 1)  # (out, in + 1, kh, kw)
        g = state.iggt.mm(g.flatten(1)).view_as(g) / state.num_locations  # (out, in + 1, kh, kw)

        if gb is not None:
            gb = g[:, -1, s[2] // 2, s[3] // 2].contiguous()  # (out)
            g = g[:, :-1]  # (out, in, kh, kw)
        g = g.contiguous()
        return g, gb

    def precond_nosua(self, mod: LayerType, weight_grad: torch.Tensor,
                      bias_grad: Optional[torch.Tensor]
                      ) -> tuple[torch.Tensor, Optional[torch.Tensor]]:
        state = self.state_storage[mod]
        g, gb = weight_grad, bias_grad
        g = g.flatten(1)  # (out, in * kh * kw)
        if gb is not None:
            g = torch.cat([g, gb.unsqueeze(1)], dim=1)  # (out, in * kh * kw + 1)
        g = state.iggt.mm(g).mm(state.ixxt) / state.num_locations  # (out, in * kh * kw + 1)
        if gb is not None:
            gb = g[:, -1].contiguous()  # (out)
            g = g[:, :-1]  # (out, in * kh * kw)
        g = g.view_as(weight_grad).contiguous()  # (out, in, kh, kw)
        return g, gb

    def compute_covs(self, mod: LayerType):
        """Computes the covariances."""
        state = self.state_storage[mod]
        x = state.x   # (N, in, xh, xw)
        gy = state.gy   # (N, out, yh, yw)

        if isinstance(mod, nn.Conv2d):
            state.num_locations = gy.size(2) * gy.size(3)  # yh * yw
            if not self.sua:
                x = F.unfold(x, mod.kernel_size,
                             padding=mod.padding,
                             stride=mod.stride)   # (N, in * kh * kw, xh * xw)
            x = x.transpose(0, 1).flatten(1)   # (in [* kh * kw], N * xh * xw)
            gy = gy.transpose(0, 1).flatten(1)   # (out, N * yh * yw)
        else:
            state.num_locations = 1
            x.t_()   # (in, N)
            gy.t_()   # (out, N)

        if mod.bias is not None:
            ones = torch.ones_like(x[:1])
            x = torch.cat([x, ones])   # (in [* kh * kw] + 1, N * xh * xw)
        if self._iteration_counter == 0:
            state.xxt = x.mm(x.t()) / x.size(1)   # (in [* kh * kw] + 1, in [* kh * kw] + 1)
            state.ggt = gy.mm(gy.t()) / gy.size(1)   # (out, out)
        else:
            state.xxt.addmm_(mat1=x, mat2=x.t(),
                             beta=(1. - self.alpha),
                             alpha=self.alpha / x.size(1))
            state.ggt.addmm_(mat1=gy, mat2=gy.t(),
                             beta=(1. - self.alpha),
                             alpha=self.alpha / gy.size(1))
        # state.xxt = (state.xxt + state.xxt.t()) / 2
        # state.ggt = (state.ggt + state.ggt.t()) / 2
        state.x = None
        state.gy = None

    @staticmethod
    def inverse(matrix: torch.Tensor, eps: float = 0.1, pi: float = 1.0) -> torch.Tensor:
        diag = math.sqrt(eps * pi) * matrix.new_ones(matrix.size(0))
        return (matrix + torch.diag(diag)).inverse()

    @classmethod
    def inv_covs(cls, xxt: torch.Tensor, ggt: torch.Tensor, num_locations: int = 1,
                 eps: float = 0.1, pi: bool = False) -> tuple[torch.Tensor, torch.Tensor]:
        """Inverses the covariances."""
        # Computes pi
        pi_value = 1.0
        if pi:
            tx = xxt.trace() * ggt.size(0)
            tg = ggt.trace() * xxt.size(0)
            pi_value = float(tx / tg)
        # Regularizes and inverse
        eps /= num_locations
        ixxt = cls.inverse(xxt, eps=eps, pi=pi_value)
        iggt = cls.inverse(ggt, eps=eps, pi=pi_value)
        # ixxt, iggt = (ixxt + ixxt.t()) / 2, (iggt + iggt.t()) / 2
        return ixxt, iggt



#!/usr/bin/env python3

from autovul.base.utils import output_memory
from autovul.base.utils.process import Process
from autovul.base.utils.output import prints

import torch
from abc import ABC, abstractmethod

from typing import TYPE_CHECKING
from collections.abc import Callable    # TODO: python 3.10
from typing import Optional, Union
if TYPE_CHECKING:
    pass


class Optimizer(ABC, Process):
    name: str = 'optimizer'

    def __init__(self, iteration: int = 20, stop_threshold: Optional[float] = None,
                 loss_fn: Callable[..., torch.Tensor] = None, **kwargs):
        super().__init__(**kwargs)
        self.param_list['optimize'] = ['iteration', 'stop_threshold']

        self.iteration = iteration
        self.stop_threshold = stop_threshold
        self.loss_fn = loss_fn

    # ----------------------Overload---------------------------------- #

    def optimize(self, _input: torch.Tensor,
                 iteration: int = None,
                 loss_fn: Callable[..., torch.Tensor] = None,
                 stop_threshold: float = None,
                 output: Union[int, list[str]] = None,
                 *args, **kwargs) -> tuple[torch.Tensor, torch.Tensor]:
        # ------------------------------ Parameter Initialization ---------------------------------- #
        iteration = iteration if iteration is not None else self.iteration
        loss_fn = loss_fn if loss_fn is not None else self.loss_fn
        stop_threshold = stop_threshold if stop_threshold is not None else self.stop_threshold
        output = self.get_output(output)
        kwargs.update(iteration=iteration,
                      loss_fn=loss_fn,
                      stop_threshold=stop_threshold,
                      output=output)
        # ----------------------------------------------------------------------------------------- #
        org_input = _input.clone().detach()
        adv_input = _input.clone().detach()
        adv_input = self.preprocess_input(adv_input=adv_input, org_input=org_input, *args, **kwargs)
        iter_list: torch.Tensor = torch.zeros(len(adv_input), dtype=torch.long)
        current_idx = torch.arange(len(iter_list))
        if 'start' in output:
            self.output_info(org_input=org_input, adv_input=adv_input,
                             mode='start', *args, **kwargs)
        # ----------------------------------------------------------------------------------------- #
        for _iter in range(iteration):
            early_stop_result = self.early_stop_check(current_idx=current_idx,
                                                      adv_input=adv_input,
                                                      org_input=org_input,
                                                      *args, **kwargs)
            not_early_stop_result = ~early_stop_result
            current_idx = current_idx[not_early_stop_result]
            iter_list[current_idx] += 1
            if early_stop_result.all():
                if 'end' in output:
                    self.output_info(org_input=org_input, adv_input=adv_input,
                                     mode='end', *args, **kwargs)
                return adv_input.detach(), iter_list
            self.update_input(current_idx=current_idx,
                              adv_input=adv_input,
                              org_input=org_input,
                              *args, **kwargs)
            if 'middle' in output:
                self.output_info(org_input=org_input, adv_input=adv_input,
                                 mode='middle', _iter=_iter, *args, **kwargs)
        early_stop_result = self.early_stop_check(current_idx=current_idx,
                                                  adv_input=adv_input,
                                                  org_input=org_input,
                                                  *args, **kwargs)
        current_idx = current_idx[~early_stop_result]
        iter_list[current_idx] += 1
        if 'end' in output:
            self.output_info(org_input=org_input, adv_input=adv_input,
                             mode='end', *args, **kwargs)
        return adv_input.detach(), torch.where(iter_list <= iteration, iter_list, -torch.ones_like(iter_list))

    def early_stop_check(self, current_idx: torch.Tensor = None,
                         adv_input: torch.Tensor = None,
                         loss_value: torch.Tensor = None,
                         loss_fn: Callable[[torch.Tensor], torch.Tensor] = None,
                         stop_threshold: float = None,
                         loss_kwargs: dict[str, torch.Tensor] = {},
                         *args, **kwargs) -> torch.Tensor:
        stop_threshold = stop_threshold if stop_threshold is not None else self.stop_threshold
        if stop_threshold is None:
            return torch.zeros(len(current_idx), dtype=torch.bool)
        if loss_value is None:
            with torch.no_grad():
                current_loss_kwargs = {k: v[current_idx] for k, v in loss_kwargs.items()}
                loss_value = loss_fn(adv_input[current_idx], **current_loss_kwargs)
        assert loss_value.dim == 1
        if adv_input[current_idx] is not None:
            assert len(loss_value) == len(current_idx)
        return loss_value < stop_threshold

    def output_info(self, mode: str = 'start', _iter: int = 0, iteration: int = 0,
                    output: list[str] = None, indent: int = None,
                    *args, **kwargs):
        output = output if output is not None else self.output
        indent = indent if indent is not None else self.indent
        if mode in ['start', 'end']:
            prints(f'{self.name} Optimize {mode}', indent=indent)
        elif mode in ['middle']:
            self.output_iter(name=self.name, _iter=_iter, iteration=iteration, indent=indent + 4)
        if 'memory' in output:
            output_memory(indent=indent + 4)

    @abstractmethod
    def update_input(self, current_idx: torch.Tensor,
                     adv_input: torch.Tensor,
                     org_input: torch.Tensor,
                     *args, **kwargs):
        ...

    def preprocess_input(self, adv_input: torch.Tensor,
                         *args, **kwargs) -> torch.Tensor:
        return adv_input



#!/usr/bin/env python3

from autovul.vision.models.imagemodel import _ImageModel, ImageModel
from autovul.vision.datasets import ImageSet
import autovul.vision.utils.model_archs.resnet_ap as resnet_ap

import torch
import torch.nn as nn
import torchvision.models
from torchvision.models.resnet import model_urls as urls
from torchvision.models.resnet import conv3x3
from collections import OrderedDict
from typing import Union


class _ResNet(_ImageModel):

    def __init__(self, name: str = 'resnet18', dataset: ImageSet = None,
                 data_shape: list[int] = None, **kwargs):
        super().__init__(name=name, dataset=dataset, data_shape=data_shape, **kwargs)
        if data_shape is None:
            assert isinstance(dataset, ImageSet)
            data_shape = dataset.data_shape
        module_list: list[nn.Module] = []
        if 's' in name.split('_'):
            from autovul.vision.utils.model_archs.resnet_s import resnet_s
            _model = resnet_s(nclasses=self.num_classes)
            module_list.append(('conv1', _model.conv1))
            module_list.append(('bn1', _model.bn1))
            module_list.append(('relu', nn.ReLU(inplace=True)))
            self.classifier = nn.Sequential(OrderedDict([
                ('fc', _model.fc)
            ]))
        else:
            model_class = name.replace('_ap', '').replace('_comp', '').replace('_s', '')
            module = resnet_ap if 'ap' in name else torchvision.models
            ModelClass = getattr(module, model_class)
            kwargs = {'pool_size': data_shape[1] // 8} if 'ap' in name else {}
            _model: Union[resnet_ap.ResNet_AP, torchvision.models.ResNet] = ModelClass(
                num_classes=self.num_classes, **kwargs)
            if 'comp' in name:
                conv1: nn.Conv2d = _model.conv1
                _model.conv1 = conv3x3(conv1.in_channels, conv1.out_channels)
                if 'resnext' in name:
                    _model.fc = nn.Linear(_model.fc.in_features // 2,
                                          _model.fc.out_features,
                                          bias=_model.fc.bias is None)
                module_list.append(('conv1', _model.conv1))
                module_list.append(('bn1', _model.bn1))
                module_list.append(('relu', _model.relu))
            else:
                module_list.append(('conv1', _model.conv1))
                module_list.append(('bn1', _model.bn1))
                module_list.append(('relu', _model.relu))
                module_list.append(('maxpool', _model.maxpool))
            self.pool = _model.avgpool  # nn.AdaptiveAvgPool2d((1, 1))
            self.classifier = nn.Sequential(OrderedDict([
                ('fc', _model.fc)  # nn.Linear(512 * block.expansion, num_classes)
            ]))
            # block.expansion = 1 if BasicBlock and 4 if Bottleneck
            # ResNet 18,34 use BasicBlock, 50 and higher use Bottleneck
        module_list.extend([('layer1', _model.layer1),
                            ('layer2', _model.layer2),
                            ('layer3', _model.layer3)])
        if not ('comp' in name and 'resnext' in name):
            module_list.append(('layer4', _model.layer4))
        self.features = nn.Sequential(OrderedDict(module_list))


class ResNet(ImageModel):
    available_models = ['resnet', 'resnet_comp', 'resnet_s',
                        'resnet18', 'resnet34', 'resnet50', 'resnet101', 'resnet152',
                        'resnet18_comp', 'resnet34_comp', 'resnet50_comp', 'resnet101_comp', 'resnet152_comp',
                        'resnet18_s', 'resnet34_s', 'resnet50_s', 'resnet101_s', 'resnet152_s',
                        'resnext50_32x4d', 'resnext101_32x8d', 'wide_resnet50_2', 'wide_resnet101_2',
                        'resnext50_32x4d_comp', 'resnext101_32x8d_comp', 'wide_resnet50_2_comp', 'wide_resnet101_2_comp',
                        'resnet18_ap_comp']

    model_urls = urls

    def __init__(self, name: str = 'resnet', layer: int = 18,
                 model: type[_ResNet] = _ResNet,
                 **kwargs):
        super().__init__(name=name, layer=layer, model=model, **kwargs)

    def get_official_weights(self, **kwargs) -> OrderedDict[str, torch.Tensor]:
        _dict = super().get_official_weights(**kwargs)
        new_dict = OrderedDict()
        for i, (key, value) in enumerate(_dict.items()):
            prefix = 'features.' if i < len(_dict) - 2 else 'classifier.'
            new_dict[prefix + key] = value
        return new_dict

    @classmethod
    def get_name(cls, name: str, layer: int = None) -> str:
        prefix = ''
        if name.startswith('wide_'):
            prefix = 'wide_'
            name = name[5:]
        return prefix + super().get_name(name, layer=layer)



#!/usr/bin/env python3
from autovul.vision.datasets.imageset import ImageSet
from autovul.vision.models.imagemodel import _ImageModel, ImageModel

import torch
import torch.nn as nn
from collections import OrderedDict

import argparse
from typing import Any


class _NATSbench(_ImageModel):

    def __init__(self, network: nn.Module = None, **kwargs):
        super().__init__(**kwargs)
        _model = network
        self.features = nn.Sequential(OrderedDict([
            ('stem', _model.stem),
            ('cells', nn.Sequential(*_model.cells)),
            ('lastact', _model.lastact),
        ]))
        self.classifier = nn.Sequential(OrderedDict([
            ('fc', _model.classifier)
        ]))


class NATSbench(ImageModel):
    available_models = ['natsbench']

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--model_index', type=int, required=True)
        group.add_argument('--model_seed', type=int)
        group.add_argument('--nats_path')
        group.add_argument('--autodl_path')
        group.add_argument('--search_space')
        return group

    def __init__(self, name: str = 'natsbench', model: type[_NATSbench] = _NATSbench,
                 model_index: int = None, model_seed: int = None,
                 dataset: ImageSet = None, dataset_name: str = None,
                 nats_path: str = '/data/rbp5354/nats/NATS-tss-v1_0-3ffb9-full',
                 autodl_path: str = '/home/rbp5354/workspace/XAutoDL/lib',
                 search_space: str = 'tss', **kwargs):
        try:
            import sys
            sys.path.append(autodl_path)
            from nats_bench import create   # type: ignore
            from models import get_cell_based_tiny_net   # type: ignore
        except ImportError as e:
            print('You need to install nats_bench and auto-dl library')
            raise e

        if dataset is not None:
            assert isinstance(dataset, ImageSet)
            kwargs['dataset'] = dataset
            if dataset_name is None:
                dataset_name = dataset.name
        assert dataset_name is not None
        self.dataset_name = dataset_name

        self.search_space = search_space
        self.model_index = model_index
        self.model_seed = model_seed

        self.api = create(nats_path, search_space, fast_mode=True, verbose=False)
        config: dict[str, Any] = self.api.get_net_config(model_index, dataset_name)
        network: nn.Module = get_cell_based_tiny_net(config)
        super().__init__(name=name, model=model, network=network, **kwargs)
        self.param_list['natsbench'] = ['model_index', 'model_seed', 'search_space']

    def get_official_weights(self, hp: str = '200', **kwargs) -> OrderedDict[str, torch.Tensor]:
        _dict: OrderedDict[str, torch.Tensor] = next(iter(self.api.get_net_param(
            self.model_index, self.dataset_name, self.model_seed, hp=hp).values()))
        new_dict: OrderedDict[str, torch.Tensor] = OrderedDict()
        for k, v in _dict.items():
            if k.startswith('stem') or k.startswith('cells') or k.startswith('lastact'):
                new_dict['features.' + k] = v
            elif k.startswith('classifier'):
                new_dict['classifier.fc' + k[10:]] = v
        return new_dict

# test

# import torch
# import torch.nn as nn
# from typing import Any

# nats_path: str = '/data/rbp5354/nats/NATS-tss-v1_0-3ffb9-full'
# autodl_path: str = '/home/rbp5354/workspace/XAutoDL/lib'
# search_space: str = 'tss'
# import sys
# sys.path.append(autodl_path)
# from models import get_cell_based_tiny_net   # type: ignore
# from nats_bench import create   # type: ignore

# api = create(nats_path, search_space, fast_mode=True, verbose=False)

# if __name__ == '__main__':

#     config: dict[str, Any] = api.get_net_config(0, 'cifar10')   # 0-15624
#     network: nn.Module = get_cell_based_tiny_net(config)



#!/usr/bin/env python3
from autovul.vision.models.imagemodel import _ImageModel, ImageModel

import torch.nn as nn
import torchvision.models


cifar10_inverted_residual_setting = [
    # t, c, n, s
    [1, 16, 1, 1],
    [6, 24, 2, 1],    # Stride 2 -> 1 for CIFAR-10
    [6, 32, 3, 2],
    [6, 64, 4, 2],
    [6, 96, 3, 1],
    [6, 160, 3, 2],
    [6, 320, 1, 1],
]


class _MobileNet(_ImageModel):

    def __init__(self, name: str = 'mobilenet_v2', **kwargs):
        try:
            sub_type: str = name[10:]
            assert sub_type in ['v2', 'v2_comp', 'v3_small', 'v3_large', 'v3_small_comp', 'v3_large_comp'], f'{name=}'
        except Exception as e:
            raise AssertionError("model name should be in ['mobilenet_v2', 'mobilenet_v2_comp', "
                                 "'mobilenet_v3_small', 'mobilenet_v3_large', "
                                 "'mobilenet_v3_small_comp', 'mobilenet_v3_large_comp']")
        super().__init__(**kwargs)
        if 'v2' in sub_type:
            inverted_residual_setting = cifar10_inverted_residual_setting if 'comp' in sub_type else None
            _model = torchvision.models.mobilenet.mobilenet_v2(num_classes=self.num_classes,
                                                               inverted_residual_setting=inverted_residual_setting)
        elif 'v3_large' in sub_type:
            _model = torchvision.models.mobilenet.mobilenet_v3_large(num_classes=self.num_classes)
        else:
            assert 'v3_small' in sub_type
            _model = torchvision.models.mobilenet.mobilenet_v3_small(num_classes=self.num_classes)
        if 'comp' in sub_type:
            conv1: nn.Conv2d = _model.features[0][0]
            _model.features[0][0] = nn.Conv2d(conv1.in_channels, conv1.out_channels, kernel_size=conv1.kernel_size,
                                              stride=1, padding=conv1.padding, dilation=conv1.dilation,
                                              groups=conv1.groups, bias=False)
        self.features = _model.features
        self.classifier = _model.classifier


class MobileNet(ImageModel):
    available_models = ['mobilenet_v2', 'mobilenet_v3_large', 'mobilenet_v3_small',
                        'mobilenet_v2_comp', 'mobilenet_v3_large_comp', 'mobilenet_v3_small_comp']

    model_urls = {
        'mobilenet_v2': 'https://download.pytorch.org/models/mobilenet_v2-b0353104.pth',
        'mobilenet_v3_large': 'https://download.pytorch.org/models/mobilenet_v3_large-8738ca79.pth',
        'mobilenet_v3_small': 'https://download.pytorch.org/models/mobilenet_v3_small-047dcff4.pth',
    }

    def __init__(self, name: str = 'mobilenet_v2', model: type[_MobileNet] = _MobileNet, **kwargs):
        super().__init__(name=name, model=model, **kwargs)



#!/usr/bin/env python3

from autovul.vision.datasets.imageset import ImageSet
import autovul.vision.utils.datasets.downsampled_imagenet as di
import argparse


class DownsampledImageNet(ImageSet):
    name = 'downsampled_imagenet'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--num_classes', type=int, help='number of classes')
        return group

    def __init__(self, norm_par: dict[str, list[float]] = {
            'mean': [122.68 / 255, 116.66 / 255, 104.01 / 255],
            'std': [63.22 / 255, 61.26 / 255, 65.09 / 255], },
            num_classes: int = 1000, **kwargs):
        self.num_classes = num_classes
        super().__init__(norm_par=norm_par, **kwargs)

    def initialize(self):
        raise NotImplementedError(
            '\n\n'
            'You need to visit \'https://image-net.org/download-images.php\' '
            'to download downsampled image data (raw format).\n'
            'There are direct links to files, but not legal to distribute. '
            'Please apply for access permission and find links yourself.\n\n'
            f'folder_path: {self.folder_path}\n'
            'expected files:\n'
            '{folder_path}/train_data_batch_*\n'
            '{folder_path}/val_data')


class ImageNet16(DownsampledImageNet):
    name = 'imagenet16'
    data_shape = [3, 16, 16]

    def _get_org_dataset(self, mode: str, **kwargs):
        assert mode in ['train', 'valid']
        return di.ImageNet16(root=self.folder_path, train=(mode == 'train'),
                             num_classes=self.num_classes if self.num_classes < 1000 else None, **kwargs)


class ImageNet32(DownsampledImageNet):
    name = 'imagenet32'
    data_shape = [3, 32, 32]

    def _get_org_dataset(self, mode: str, **kwargs):
        assert mode in ['train', 'valid']
        return di.ImageNet32(root=self.folder_path, train=(mode == 'train'),
                             num_classes=self.num_classes if self.num_classes < 1000 else None, **kwargs)



#!/usr/bin/env python3

# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3

# https://github.com/google-research/big_transfer/blob/master/bit_pytorch/models.py
"""Bottleneck ResNet v2 with GroupNorm and Weight Standardization."""

import torch
import torch.nn as nn
import torch.nn.utils.parametrize as parametrize
import numpy as np
from collections import OrderedDict


class Std(nn.Module):
    def forward(self, X: torch.Tensor):
        v, m = torch.var_mean(X, dim=[1, 2, 3], keepdim=True, unbiased=False)
        return (X - m) / torch.sqrt(v + 1e-10)


class StdConv2d(nn.Conv2d):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        parametrize.register_parametrization(self, "weight", Std())


# class StdConv2d(nn.Conv2d):
#     def forward(self, x: torch.Tensor) -> torch.Tensor:
#         w = self.weight
#         v, m = torch.var_mean(w, dim=[1, 2, 3], keepdim=True, unbiased=False)
#         w = (w - m) / torch.sqrt(v + 1e-10)
#         return F.conv2d(x, w, self.bias, self.stride, self.padding, self.dilation, self.groups)
#     # TODO: Linting for __call__


def conv3x3(cin: int, cout: int, stride: int = 1, groups: int = 1, bias: bool = False):
    return StdConv2d(cin, cout, kernel_size=3, stride=stride, padding=1, bias=bias, groups=groups)


def conv1x1(cin: int, cout: int, stride: int = 1, bias: bool = False):
    return StdConv2d(cin, cout, kernel_size=1, stride=stride, padding=0, bias=bias)


def tf2th(conv_weights: np.ndarray) -> torch.Tensor:
    """Possibly convert HWIO to OIHW"""
    if conv_weights.ndim == 4:
        conv_weights = np.transpose(conv_weights, [3, 2, 0, 1])
    return torch.from_numpy(conv_weights)


class PreActBottleneck(nn.Module):
    """Pre-activation (v2) bottleneck block.
    Follows the implementation of "Identity Mappings in Deep Residual Networks":
    https://github.com/KaimingHe/resnet-1k-layers/blob/master/resnet-pre-act.lua
    Except it puts the stride on 3x3 conv when available.
    """

    def __init__(self, cin: int, cout: int = None, cmid: int = None, stride: int = 1):
        super().__init__()
        cout = cout or cin
        cmid = cmid or cout // 4

        self.gn1 = nn.GroupNorm(32, cin)
        self.conv1 = conv1x1(cin, cmid)
        self.gn2 = nn.GroupNorm(32, cmid)
        self.conv2 = conv3x3(cmid, cmid, stride)  # Original code has it on conv1!!
        self.gn3 = nn.GroupNorm(32, cmid)
        self.conv3 = conv1x1(cmid, cout)
        self.relu = nn.ReLU(inplace=True)

        if (stride != 1 or cin != cout):
            # Projection also with pre-activation according to paper.
            self.downsample = conv1x1(cin, cout, stride)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        out = self.relu(self.gn1(x))

        # Residual branch
        residual = x
        if hasattr(self, 'downsample'):
            residual = self.downsample(out)

        # Unit's branch
        out = self.conv1(out)
        out = self.conv2(self.relu(self.gn2(out)))
        out = self.conv3(self.relu(self.gn3(out)))

        return out + residual

    def load_from(self, weights: dict[str, np.ndarray], prefix: str = ''):
        convname = 'standardized_conv2d'
        with torch.no_grad():
            self.conv1.weight.copy_(tf2th(weights[f'{prefix}a/{convname}/kernel']))
            self.conv2.weight.copy_(tf2th(weights[f'{prefix}b/{convname}/kernel']))
            self.conv3.weight.copy_(tf2th(weights[f'{prefix}c/{convname}/kernel']))
            self.gn1.weight.copy_(tf2th(weights[f'{prefix}a/group_norm/gamma']))
            self.gn2.weight.copy_(tf2th(weights[f'{prefix}b/group_norm/gamma']))
            self.gn3.weight.copy_(tf2th(weights[f'{prefix}c/group_norm/gamma']))
            self.gn1.bias.copy_(tf2th(weights[f'{prefix}a/group_norm/beta']))
            self.gn2.bias.copy_(tf2th(weights[f'{prefix}b/group_norm/beta']))
            self.gn3.bias.copy_(tf2th(weights[f'{prefix}c/group_norm/beta']))
            if hasattr(self, 'downsample'):
                w = weights[f'{prefix}a/proj/{convname}/kernel']
                self.downsample.weight.copy_(tf2th(w))


class ResNetV2(nn.Module):
    """Implementation of Pre-activation (v2) ResNet mode."""

    def __init__(self, block_units: list[int], width_factor: int, head_size: int = 21843, zero_head: bool = False):
        super().__init__()
        wf = width_factor  # shortcut 'cause we'll use it a lot.

        # The following will be unreadable if we split lines.
        # pylint: disable=line-too-long
        self.root = nn.Sequential(OrderedDict([
            ('conv', StdConv2d(3, 64 * wf, kernel_size=7, stride=2, padding=3, bias=False)),
            ('pad', nn.ConstantPad2d(1, 0)),
            ('pool', nn.MaxPool2d(kernel_size=3, stride=2, padding=0)),
            # The following is subtly not the same!
            # ('pool', nn.MaxPool2d(kernel_size=3, stride=2, padding=1)),
        ]))

        self.body = nn.Sequential(OrderedDict([
            ('block1', nn.Sequential(OrderedDict(
                [('unit01', PreActBottleneck(cin=64 * wf, cout=256 * wf, cmid=64 * wf))] +
                [(f'unit{i:02d}', PreActBottleneck(cin=256 * wf, cout=256 * wf, cmid=64 * wf))
                 for i in range(2, block_units[0] + 1)],
            ))),
            ('block2', nn.Sequential(OrderedDict(
                [('unit01', PreActBottleneck(cin=256 * wf, cout=512 * wf, cmid=128 * wf, stride=2))] +
                [(f'unit{i:02d}', PreActBottleneck(cin=512 * wf, cout=512 * wf, cmid=128 * wf))
                 for i in range(2, block_units[1] + 1)],
            ))),
            ('block3', nn.Sequential(OrderedDict(
                [('unit01', PreActBottleneck(cin=512 * wf, cout=1024 * wf, cmid=256 * wf, stride=2))] +
                [(f'unit{i:02d}', PreActBottleneck(cin=1024 * wf, cout=1024 * wf, cmid=256 * wf))
                 for i in range(2, block_units[2] + 1)],
            ))),
            ('block4', nn.Sequential(OrderedDict(
                [('unit01', PreActBottleneck(cin=1024 * wf, cout=2048 * wf, cmid=512 * wf, stride=2))] +
                [(f'unit{i:02d}', PreActBottleneck(cin=2048 * wf, cout=2048 * wf, cmid=512 * wf))
                 for i in range(2, block_units[3] + 1)],
            ))),
        ]))
        # pylint: enable=line-too-long

        self.zero_head = zero_head
        self.head = nn.Sequential(OrderedDict([
            ('gn', nn.GroupNorm(32, 2048 * wf)),
            ('relu', nn.ReLU(inplace=True)),
            ('avg', nn.AdaptiveAvgPool2d(output_size=1)),
            ('conv', nn.Conv2d(2048 * wf, head_size, kernel_size=1, bias=True)),
        ]))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.head(self.body(self.root(x)))
        assert x.shape[-2:] == (1, 1)  # We should have no spatial shape left.
        return x[..., 0, 0]

    def load_from(self, weights: dict[str, np.ndarray], prefix: str = 'resnet/'):
        with torch.no_grad():
            self.root.conv.weight.copy_(
                tf2th(weights[f'{prefix}root_block/standardized_conv2d/kernel']))  # pylint: disable=line-too-long
            self.head.gn.weight.copy_(tf2th(weights[f'{prefix}group_norm/gamma']))
            self.head.gn.bias.copy_(tf2th(weights[f'{prefix}group_norm/beta']))
            if self.zero_head:
                nn.init.zeros_(self.head.conv.weight)
                nn.init.zeros_(self.head.conv.bias)
            else:
                self.head.conv.weight.copy_(
                    tf2th(weights[f'{prefix}head/conv2d/kernel']))  # pylint: disable=line-too-long
                self.head.conv.bias.copy_(tf2th(weights[f'{prefix}head/conv2d/bias']))

            for bname, block in self.body.named_children():
                for uname, unit in block.named_children():
                    unit.load_from(weights, prefix=f'{prefix}{bname}/{uname}/')


KNOWN_MODELS = OrderedDict([
    ('BiT-M-R50x1', lambda *a, **kw: ResNetV2([3, 4, 6, 3], 1, *a, **kw)),
    ('BiT-M-R50x3', lambda *a, **kw: ResNetV2([3, 4, 6, 3], 3, *a, **kw)),
    ('BiT-M-R101x1', lambda *a, **kw: ResNetV2([3, 4, 23, 3], 1, *a, **kw)),
    ('BiT-M-R101x3', lambda *a, **kw: ResNetV2([3, 4, 23, 3], 3, *a, **kw)),
    ('BiT-M-R152x2', lambda *a, **kw: ResNetV2([3, 8, 36, 3], 2, *a, **kw)),
    ('BiT-M-R152x4', lambda *a, **kw: ResNetV2([3, 8, 36, 3], 4, *a, **kw)),
    ('BiT-S-R50x1', lambda *a, **kw: ResNetV2([3, 4, 6, 3], 1, *a, **kw)),
    ('BiT-S-R50x3', lambda *a, **kw: ResNetV2([3, 4, 6, 3], 3, *a, **kw)),
    ('BiT-S-R101x1', lambda *a, **kw: ResNetV2([3, 4, 23, 3], 1, *a, **kw)),
    ('BiT-S-R101x3', lambda *a, **kw: ResNetV2([3, 4, 23, 3], 3, *a, **kw)),
    ('BiT-S-R152x2', lambda *a, **kw: ResNetV2([3, 8, 36, 3], 2, *a, **kw)),
    ('BiT-S-R152x4', lambda *a, **kw: ResNetV2([3, 8, 36, 3], 4, *a, **kw)),
])



#!/usr/bin/env python3

from .imageset import ImageSet
from .imagefolder import ImageFolder

from .folder import *
from .normal import *

from . import folder, normal

from autovul.vision.configs import Config, config
import autovul.base.datasets

import argparse
from typing import Union

module_list = [folder, normal]
__all__ = ['ImageSet', 'ImageFolder', 'class_dict', 'add_argument', 'create']
class_dict: dict[str, type[ImageSet]] = {}
for module in module_list:
    __all__.extend(module.__all__)
    class_dict.update(module.class_dict)


def add_argument(parser: argparse.ArgumentParser, dataset_name: str = None, dataset: Union[str, ImageSet] = None,
                 config: Config = config, class_dict: dict[str, type[ImageSet]] = class_dict) -> argparse._ArgumentGroup:
    return autovul.base.datasets.add_argument(parser=parser, dataset_name=dataset_name, dataset=dataset,
                                              config=config, class_dict=class_dict)


def create(dataset_name: str = None, dataset: str = None,
           config: Config = config, class_dict: dict[str, type[ImageSet]] = class_dict, **kwargs) -> ImageSet:
    return autovul.base.datasets.create(dataset_name=dataset_name, dataset=dataset,
                                        config=config, class_dict=class_dict, **kwargs)



#!/usr/bin/env python3

from autovul.vision.datasets.imageset import ImageSet
from autovul.vision.models.imagemodel import ImageModel
from autovul.vision.marks import Watermark
from autovul.base.attacks import Attack
from autovul.base.utils import to_list
from autovul.base.utils.data import TensorListDataset, dataset_to_list
from autovul.base.utils.output import prints
from autovul.base.utils.logger import SmoothedValue


import torch
import math
import random
import os
import argparse
from typing import Callable
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    import torch.utils.data


class BadNet(Attack):
    r"""
    BadNet Backdoor Attack is described in detail in the paper `BadNet`_ by Tianyu Gu. 

    It attaches a fixed watermark to benign images and inject them into training set with target label.
    After retraining, the model will classify all images with watermark attached into target class.

    The authors have posted `original source code`_.

    Args:
        mark (Watermark): the attached watermark image.
        target_class (int): the target class. Default: ``0``.
        poison_percent (int): The proportion of malicious images in the training set (Max 0.5). Default: 0.1.

    .. _BadNet:
        https://arxiv.org/abs/1708.06733

    .. _original source code:
        https://github.com/Kooscii/BadNets
    """

    name: str = 'badnet'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--target_class', type=int, help='target class of backdoor, defaults to 0')
        group.add_argument('--poison_percent', type=float,
                           help='malicious training data injection probability for each batch, defaults to 0.01')
        group.add_argument('--train_mode', help='target class of backdoor, defaults to "batch"')
        return group

    def __init__(self, mark: Watermark = None, target_class: int = 0, poison_percent: float = 0.01, train_mode: str = 'batch', **kwargs):
        super().__init__(**kwargs)
        self.dataset: ImageSet
        self.model: ImageModel
        self.param_list['badnet'] = ['train_mode', 'target_class', 'poison_percent', 'poison_num']
        self.mark: Watermark = mark
        self.target_class: int = target_class
        self.poison_percent: float = poison_percent
        self.poison_num = self.dataset.batch_size * self.poison_percent
        self.train_mode: str = train_mode

    def attack(self, epoch: int, save=False, **kwargs):
        if self.train_mode == 'batch':
            self.model._train(epoch, save=save,
                              validate_fn=self.validate_fn, get_data_fn=self.get_data,
                              save_fn=self.save, **kwargs)
        elif self.train_mode == 'dataset':
            dataset = self.mix_dataset()
            loader = self.dataset.get_dataloader('train', dataset=dataset)
            self.model._train(epoch, save=save,
                              validate_fn=self.validate_fn, loader_train=loader,
                              save_fn=self.save, **kwargs)
        elif self.train_mode == 'loss':
            self.model._train(epoch, save=save,
                              validate_fn=self.validate_fn, loss_fn=self.loss_fn,
                              save_fn=self.save, **kwargs)

    def mix_dataset(self, poison_label: bool = True) -> torch.utils.data.Dataset:
        clean_dataset = self.dataset.loader['train'].dataset
        subset, _ = ImageSet.split_dataset(clean_dataset, percent=self.poison_percent)
        _input, _label = dataset_to_list(subset)
        _input = torch.stack(_input)

        if poison_label:
            _label = [self.target_class] * len(_label)
        poison_input = self.add_mark(_input)
        poison_dataset = TensorListDataset(poison_input, _label)
        return torch.utils.data.ConcatDataset([clean_dataset, poison_dataset])

    def get_filename(self, mark_alpha: float = None, target_class: int = None, **kwargs):
        if mark_alpha is None:
            mark_alpha = self.mark.mark_alpha
        if target_class is None:
            target_class = self.target_class
        mark_filename = os.path.split(self.mark.mark_path)[-1]
        mark_name, mark_ext = os.path.splitext(mark_filename)
        _file = '{mark}_tar{target:d}_alpha{mark_alpha:.2f}_mark({mark_height:d},{mark_width:d})'.format(
            mark=mark_name, target=target_class, mark_alpha=mark_alpha,
            mark_height=self.mark.mark_height, mark_width=self.mark.mark_width)
        if self.mark.random_pos:
            _file = 'random_pos_' + _file
        if self.mark.mark_distributed:
            _file = 'distributed_' + _file
        return _file

    # ---------------------- I/O ----------------------------- #

    def save(self, filename: str = None, **kwargs):
        filename = filename if filename is not None else self.get_filename(**kwargs)
        file_path = os.path.join(self.folder_path, filename)
        self.mark.save_npz(file_path + '.npz')
        self.mark.save_img(file_path + '.png')
        self.model.save(file_path + '.pth')
        print('attack results saved at: ', file_path)

    def load(self, filename: str = None, **kwargs):
        filename = filename if filename is not None else self.get_filename(**kwargs)
        file_path = os.path.join(self.folder_path, filename)
        self.mark.load_npz(file_path + '.npz')
        self.model.load(file_path + '.pth')
        print('attack results loaded from: ', file_path)

    # ---------------------- Utils ---------------------------- #

    def add_mark(self, x: torch.Tensor, **kwargs) -> torch.Tensor:
        return self.mark.add_mark(x, **kwargs)

    def loss_fn(self, _input: torch.Tensor = None, _label: torch.Tensor = None, _output: torch.Tensor = None, **kwargs) -> torch.Tensor:
        loss_clean = self.model.loss(_input, _label, **kwargs)
        poison_input = self.mark.add_mark(_input)
        poison_label = self.target_class * torch.ones_like(_label)
        loss_poison = self.model.loss(poison_input, poison_label, **kwargs)
        return (1 - self.poison_percent) * loss_clean + self.poison_percent * loss_poison

    def get_data(self, data: tuple[torch.Tensor, torch.Tensor], keep_org: bool = True,
                 poison_label=True, **kwargs) -> tuple[torch.Tensor, torch.Tensor]:
        _input, _label = self.model.get_data(data)
        decimal, integer = math.modf(self.poison_num)
        integer = int(integer)
        if random.uniform(0, 1) < decimal:
            integer += 1
        if not keep_org:
            integer = len(_label)
        if not keep_org or integer:
            org_input, org_label = _input, _label
            _input = self.add_mark(org_input[:integer])
            _label = _label[:integer]
            if poison_label:
                _label = self.target_class * torch.ones_like(org_label[:integer])
            if keep_org:
                _input = torch.cat((_input, org_input))
                _label = torch.cat((_label, org_label))
        return _input, _label

    def validate_fn(self,
                    get_data_fn: Callable[..., tuple[torch.Tensor, torch.Tensor]] = None,
                    loss_fn: Callable[..., torch.Tensor] = None,
                    main_tag: str = 'valid', indent: int = 0, **kwargs) -> tuple[float, float]:
        _, clean_acc = self.model._validate(print_prefix='Validate Clean', main_tag='valid clean',
                                            get_data_fn=None, indent=indent, **kwargs)
        _, target_acc = self.model._validate(print_prefix='Validate Trigger Tgt', main_tag='valid trigger target',
                                             get_data_fn=self.get_data, keep_org=False, poison_label=True,
                                             indent=indent, **kwargs)
        self.model._validate(print_prefix='Validate Trigger Org', main_tag='',
                             get_data_fn=self.get_data, keep_org=False, poison_label=False,
                             indent=indent, **kwargs)
        prints(f'Validate Confidence: {self.validate_confidence():.3f}', indent=indent)
        prints(f'Neuron Jaccard Idx: {self.check_neuron_jaccard():.3f}', indent=indent)
        if self.clean_acc - clean_acc > 3 and self.clean_acc > 40:  # TODO: better not hardcoded
            target_acc = 0.0
        return clean_acc, target_acc

    def validate_confidence(self) -> float:
        confidence = SmoothedValue()
        with torch.no_grad():
            for data in self.dataset.loader['valid']:
                _input, _label = self.model.get_data(data)
                idx1 = _label != self.target_class
                _input = _input[idx1]
                _label = _label[idx1]
                if len(_input) == 0:
                    continue
                poison_input = self.add_mark(_input)
                poison_label = self.model.get_class(poison_input)
                idx2 = poison_label == self.target_class
                poison_input = poison_input[idx2]
                if len(poison_input) == 0:
                    continue
                batch_conf = self.model.get_prob(poison_input)[:, self.target_class].mean()
                confidence.update(batch_conf, len(poison_input))
        return confidence.global_avg

    def check_neuron_jaccard(self, ratio=0.5) -> float:
        feats_list = []
        poison_feats_list = []
        with torch.no_grad():
            for data in self.dataset.loader['valid']:
                _input, _label = self.model.get_data(data)
                poison_input = self.add_mark(_input)

                _feats = self.model.get_final_fm(_input)
                poison_feats = self.model.get_final_fm(poison_input)
                feats_list.append(_feats)
                poison_feats_list.append(poison_feats)
        feats_list = torch.cat(feats_list).mean(dim=0)
        poison_feats_list = torch.cat(poison_feats_list).mean(dim=0)
        length = int(len(feats_list) * ratio)
        _idx = set(to_list(feats_list.argsort(descending=True))[:length])
        poison_idx = set(to_list(poison_feats_list.argsort(descending=True))[:length])
        jaccard_idx = len(_idx & poison_idx) / len(_idx | poison_idx)
        return jaccard_idx



#!/usr/bin/env python3

from autovul.base.datasets import Dataset
from autovul.base.configs import config, Config
from autovul.base.utils import get_name
from autovul.base.utils.output import ansi
from autovul.base.utils.process import Model_Process

import torch
import os

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING
from typing import Union    # TODO: python 3.10
from autovul.base.models import Model
import argparse    # TODO: python 3.10
if TYPE_CHECKING:
    pass


class Attack(ABC, Model_Process):
    name = 'attack'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        group.add_argument('--attack', dest='attack_name')
        group.add_argument('--attack_dir', help='directory to contain attack results')
        group.add_argument('--output', type=int, help='output level, defaults to 0.')
        return group

    @abstractmethod
    def attack(self, **kwargs):
        ...

    def generate_target(self, _input: torch.Tensor, idx: int = 1, same: bool = False, **kwargs):
        # TODO: issue 1 anyway to avoid coding so many arguments?
        return self.model.generate_target(_input, idx=idx, same=same, **kwargs)


def add_argument(parser: argparse.ArgumentParser, attack_name: str = None, attack: Union[str, Attack] = None,
                 class_dict: dict[str, type[Attack]] = {}):
    attack_name = get_name(name=attack_name, module=attack, arg_list=['--attack'])
    group = parser.add_argument_group('{yellow}attack{reset}'.format(**ansi), description=attack_name)
    try:
        AttackType = class_dict[attack_name]
    except KeyError as e:
        if attack_name is None:
            print('{red}you need to first claim the attack name using "--attack".{reset}'.format(**ansi))
        print(f'{attack_name} not in \n{list(class_dict.keys())}')
        raise e
    return AttackType.add_argument(group)


def create(attack_name: str = None, attack: Union[str, Attack] = None, folder_path: str = None,
           dataset_name: str = None, dataset: Union[str, Dataset] = None,
           model_name: str = None, model: Union[str, Model] = None,
           config: Config = config, class_dict: dict[str, type[Attack]] = {}, **kwargs):
    dataset_name = get_name(name=dataset_name, module=dataset, arg_list=['-d', '--dataset'])
    model_name = get_name(name=model_name, module=model, arg_list=['-m', '--model'])
    attack_name = get_name(name=attack_name, module=attack, arg_list=['--attack'])
    if dataset_name is None:
        dataset_name = config.get_full_config()['dataset']['default_dataset']
    general_config = config.get_config(dataset_name=dataset_name)['attack']
    specific_config = config.get_config(dataset_name=dataset_name)[attack_name]
    result = general_config.update(specific_config).update(kwargs)
    try:
        AttackType = class_dict[attack_name]
    except KeyError as e:
        print(f'{attack_name} not in \n{list(class_dict.keys())}')
        raise e
    if folder_path is None:
        folder_path = result['attack_dir']
        if isinstance(dataset, Dataset):
            folder_path = os.path.join(folder_path, dataset.data_type, dataset.name)
        if model_name is not None:
            folder_path = os.path.join(folder_path, model_name)
        folder_path = os.path.join(folder_path, AttackType.name)
    return AttackType(name=attack_name, dataset=dataset, model=model, folder_path=folder_path, **result)



#!/usr/bin/env python3

from autovul.vision.datasets.imageset import ImageSet

import torchvision.datasets as datasets


class MNIST(ImageSet):

    name: str = 'mnist'
    num_classes: int = 10
    data_shape = [1, 28, 28]

    def __init__(self, norm_par: dict[str, list[int]] = {'mean': [0.1307], 'std': [0.3081]}, **kwargs):
        super().__init__(norm_par=norm_par, **kwargs)

    def initialize(self):
        datasets.MNIST(root=self.folder_path, train=True, download=True)
        datasets.MNIST(root=self.folder_path, train=False, download=True)

    def _get_org_dataset(self, mode, **kwargs):
        assert mode in ['train', 'valid']
        return datasets.MNIST(root=self.folder_path, train=(mode == 'train'), **kwargs)



#!/usr/bin/env python3

import autovul.vision
from autovul.vision.utils import summary, to_numpy
from autovul.base.utils.data import dataset_to_list

import torch
import torch.nn as nn
import argparse


if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    parser.add_argument('--nb_stolen', type=int, dest="nb_stolen", default=5000)
    parser.add_argument('--tmodel', dest="tmodel", default=None)
    parser.add_argument('--tmodel_arch', dest="tmodel_arch", default=None)
    autovul.vision.environ.add_argument(parser)
    autovul.vision.datasets.add_argument(parser)
    autovul.vision.models.add_argument(parser)
    autovul.vision.trainer.add_argument(parser)
    args = parser.parse_args()
    nb_stolen = args.nb_stolen
    # print(nb_stolen)

    env = autovul.vision.environ.create(**args.__dict__)
    dataset = autovul.vision.datasets.create(**args.__dict__)
    model = autovul.vision.models.create(dataset=dataset, **args.__dict__)

    if env['verbose']:
        summary(env=env, dataset=dataset, model=model)
    # model._validate()
    model.eval()
    # print('\n\n')

    from art.estimators.classification import PyTorchClassifier  # type: ignore
    classifier = PyTorchClassifier(
        model=model._model,
        loss=model.criterion,
        input_shape=dataset.data_shape,
        nb_classes=model.num_classes,
    )
    x_train, y_train = dataset_to_list(dataset.get_dataset('train'))
    x_train, y_train = to_numpy(torch.stack(x_train)), to_numpy(y_train)

    import art.attacks.extraction  # type:ignore
    # for name in ['CopycatCNN', 'KnockoffNets']:
    AttackClass = getattr(art.attacks.extraction, 'KnockoffNets')
    for mode in ['random']:

        model_name = args.model_name if not args.tmodel else args.tmodel

        # print(model_name, model_arch)

        args_dict = args.__dict__ | {'pretrain': False, 'official': False,
                                     'model_name': model_name}
        if args.tmodel_arch is not None:
            args_dict['model_arch'] = args.tmodel_arch
        elif 'model_arch' in args.__dict__.keys():
            args_dict['model_arch'] = args.model_arch

        thieved_model = autovul.vision.models.create(dataset=dataset, **args_dict)
        trainer = autovul.vision.trainer.create(dataset=dataset, model=thieved_model, **args.__dict__)

        thieved_model.train()
        params: list[nn.Parameter] = []
        for param_group in trainer.optimizer.param_groups:
            params.extend(param_group['params'])
        thieved_model.activate_params(params)

        thieved_classifier = PyTorchClassifier(
            model=thieved_model._model,
            loss=thieved_model.criterion,
            input_shape=dataset.data_shape,
            nb_classes=thieved_model.num_classes,
            optimizer=trainer.optimizer
        )

        # thieved_model._validate(print_prefix='Before Stealing')
        attack = AttackClass(classifier, batch_size_fit=dataset.batch_size, batch_size_query=dataset.batch_size,
                             nb_epochs=25, nb_stolen=nb_stolen, use_probability=True, sampling_strategy=mode, verbose=False)
        attack.extract(x=x_train, y=y_train, thieved_classifier=thieved_classifier)

        thieved_model.activate_params([])
        thieved_model.eval()
        thieved_model._validate(print_prefix='KnockoffNets-' + mode)
        model._compare(thieved_model)
        print('-' * 20)
        # print('\n\n')
